import copy
import datetime
import json
from pathlib import Path
from typing import Any, Iterable, Literal, Self, cast

from anthropic.types.message_param import MessageParam

from google.genai.types import ContentUnionDict  # type: ignore
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam
from openai.types.responses.response_input_param import ResponseInputItemParam
from pydantic import BaseModel, Field, PrivateAttr, computed_field, model_validator

from ..._utils.chat import convert_to_anthropic_format, convert_to_google_genai_format
from ..._utils.chat import convert_to_openai_format as convert_to_openai_completions_api_format
from ..._utils.json_schema import (
    convert_basemodel_to_partial_basemodel,
    convert_json_schema_to_basemodel,
    create_reasoning_schema,
    expand_refs,
    generate_schema_data_id,
    generate_schema_id,
    json_schema_to_nlp_data_structure,
    json_schema_to_strict_openai_schema,
    json_schema_to_typescript_interface,
    load_json_schema,
    schema_to_ts_type,
)
from ..._utils.responses import convert_to_openai_format as convert_to_openai_responses_api_format
from ...types.standards import StreamingBaseModel
from ..chat import ChatCompletionUiformMessage


class PartialSchema(BaseModel):
    """Response from the Generate Schema API -- A partial Schema object with no validation"""

    object: Literal["schema"] = "schema"
    created_at: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(datetime.timezone.utc))
    json_schema: dict[str, Any] = {}
    strict: bool = True


class PartialSchemaChunk(StreamingBaseModel):
    object: Literal["schema.chunk"] = "schema.chunk"
    created_at: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(datetime.timezone.utc))
    delta_json_schema_flat: dict[str, Any] = {}


# class PartialSchemaStreaming(StreamingBaseModel, PartialSchema): pass


class Schema(PartialSchema):
    """A full Schema object with validation."""

    object: Literal["schema"] = "schema"
    """The type of object being preprocessed."""

    created_at: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(datetime.timezone.utc))
    """The timestamp of when the schema was created."""

    json_schema: dict[str, Any] = {}
    """The JSON schema to use for loading."""

    # This is a computed field, it is exposed when serializing the object
    @computed_field  # type: ignore
    @property
    def data_id(self) -> str:
        """Returns the SHA1 hash of the schema data, ignoring all prompt/description/default fields.

        Returns:
            str: A SHA1 hash string representing the schema data version.
        """
        return generate_schema_data_id(self.json_schema)

    # This is a computed field, it is exposed when serializing the object
    @computed_field  # type: ignore
    @property
    def id(self) -> str:
        """Returns the SHA1 hash of the complete schema.

        Returns:
            str: A SHA1 hash string representing the complete schema version.
        """
        return generate_schema_id(self.json_schema)

    pydantic_model: type[BaseModel] = Field(default=None, exclude=True, repr=False)  # type: ignore

    _partial_pydantic_model: type[BaseModel] = PrivateAttr()
    """The Pydantic model to use for loading."""

    @property
    def inference_pydantic_model(self) -> type[BaseModel]:
        """Converts the structured output schema to a Pydantic model, with the LLMDescription and ReasoningDescription fields added.

        Returns:
            type[BaseModel]: A Pydantic model class generated from the schema.
        """
        return convert_json_schema_to_basemodel(self.inference_json_schema)

    @property
    def inference_json_schema(self) -> dict[str, Any]:
        """Returns the schema formatted for structured output, with the LLMDescription and ReasoningDescription fields added.

        Returns:
            dict[str, Any]: The schema formatted for structured output processing.
        """
        if self.strict:
            inference_json_schema_ = json_schema_to_strict_openai_schema(copy.deepcopy(self._reasoning_object_schema))
            assert isinstance(inference_json_schema_, dict), "Validation Error: The inference_json_schema is not a dict"
            return inference_json_schema_
        else:
            return copy.deepcopy(self._reasoning_object_schema)

    @property
    def openai_messages(self) -> list[ChatCompletionMessageParam]:
        """Returns the messages formatted for OpenAI's API.

        Returns:
            list[ChatCompletionMessageParam]: List of messages in OpenAI's format.
        """
        return convert_to_openai_completions_api_format(self.messages)

    @property
    def openai_responses_input(self) -> list[ResponseInputItemParam]:
        """Returns the messages formatted for OpenAI's Responses API.

        Returns:
            list[ResponseInputItemParam]: List of messages in OpenAI's Responses API format.
        """
        return convert_to_openai_responses_api_format(self.messages)

    @property
    def anthropic_system_prompt(self) -> str:
        """Returns the system message in Anthropic's Claude format.

        Returns:
            str : The system prompt formatted for Claude.
        """
        return "Return your response as a JSON object following the provided schema." + self.system_prompt

    @property
    def anthropic_messages(self) -> list[MessageParam]:
        """Returns the messages in Anthropic's Claude format.

        Returns:
            list[MessageParam]: List of messages formatted for Claude.
        """
        return convert_to_anthropic_format(self.messages)[1]

    @property
    def gemini_system_prompt(self) -> str:
        return convert_to_google_genai_format(self.messages)[0]

    @property
    def gemini_messages(self) -> list[ContentUnionDict]:
        """Returns the messages formatted for Google's Gemini API."""
        return convert_to_google_genai_format(self.messages)[1]

    @property
    def inference_gemini_json_schema(self) -> dict[str, Any]:
        # Like OpenAI but does not accept "anyOf" typing, all fields must not be nullable
        inference_json_schema_ = copy.deepcopy(self._reasoning_object_schema)

        def json_schema_to_gemini_schema(schema: dict[str, Any]) -> None:
            if "$defs" in schema:
                for def_schema in schema["$defs"].values():
                    json_schema_to_gemini_schema(def_schema)
            if "anyOf" in schema:
                any_of = schema.pop("anyOf")
                is_nullable = any(s.get("type") == "null" for s in any_of)
                # Get the non-null subschemas
                non_null_schemas = [s for s in any_of if s.get("type") != "null"]

                if non_null_schemas:
                    subschema = non_null_schemas[0]
                    json_schema_to_gemini_schema(subschema)
                    # Take the first non-null subschema and merge it into the parent schema
                    schema.update(subschema)
                else:
                    raise ValueError("No non-null subschemas found within anyOf")

                if is_nullable and schema.get("type") not in ["object", "array"]:
                    schema["nullable"] = True

            if "allOf" in schema:
                for allof_schema in schema["allOf"]:
                    json_schema_to_gemini_schema(allof_schema)

            if schema.get("type") == "object" and "properties" in schema:
                for prop_schema in schema["properties"].values():
                    json_schema_to_gemini_schema(prop_schema)
                schema["propertyOrdering"] = schema["required"] = list(schema["properties"].keys())

            if schema.get("type") == "array" and "items" in schema:
                json_schema_to_gemini_schema(schema["items"])
            # Remove not allowed fields
            for key in ["additionalProperties", "format"]:
                schema.pop(key, None)

        json_schema_to_gemini_schema(inference_json_schema_)
        return inference_json_schema_

    @property
    def inference_typescript_interface(self) -> str:
        """Returns the TypeScript interface representation of the inference schema, that is more readable than the JSON schema.

        Returns:
            str: A string containing the TypeScript interface definition.
        """
        return json_schema_to_typescript_interface(self._reasoning_object_schema, add_field_description=False)

    @property
    def inference_nlp_data_structure(self) -> str:
        """Returns the NLP data structure representation of the inference schema, that is more readable than the JSON schema.

        Returns:
            str: A string containing the NLP data structure definition.
        """
        return json_schema_to_nlp_data_structure(self._reasoning_object_schema)

    @property
    def developer_system_prompt(self) -> str:
        return '''
# General Instructions

You are an expert in data extraction and structured data outputs.

When provided with a **JSON schema** and a **document**, you must:

1. Carefully extract all relevant data from the provided document according to the given schema.
2. Return extracted data strictly formatted according to the provided schema.
3. Make sure that the extracted values are **UTF-8** encodable strings.
4. Avoid generating bytes, binary data, base64 encoded data, or other non-UTF-8 encodable data.

---

## Handling Missing and Nullable Fields

### Nullable Leaf Attributes

- If valid data is missing or not explicitly present, set leaf attributes explicitly to `null`.
- **Do NOT** use empty strings (`""`), placeholder values, or fabricated data.

**Example:**

```json
// Correct:
{"email": null}

// Incorrect:
{"email": ""}
```

### Nullable Nested Objects

- If an entire nested object’s data is missing or incomplete, **do NOT** set the object itself to `null`.
- Keep the object structure fully intact, explicitly setting each leaf attribute within to `null`.
- This preserves overall structure and explicitly communicates exactly which fields lack data.

**Example:**

```json
// Correct (all information is missing):
{
  "address": {
    "street": null,
    "zipCode": null,
    "city": null
  }
}

// Incorrect (all information is missing):
{
  "address": null
}

// Correct (only some information is missing):
{
  "address": {
    "street": null,
    "zipCode": null,
    "city": "Paris"
  }
}

// Incorrect (only some information is missing):
{
  "address": {
    "city": "Paris"
  }
}
```

---

## Reasoning Fields

Your schema includes special reasoning fields (`reasoning___*`) used exclusively to document your extraction logic. These fields are for detailed explanations and will not appear in final outputs.

| Reasoning Field Type | Field Naming Pattern       |
|----------------------|----------------------------|
| Root Object          | `reasoning___root`         |
| Nested Objects       | `reasoning___[objectname]` |
| Array Fields         | `reasoning___[arrayname]`  |
| Array Elements       | `reasoning___item`         |
| Leaf Attributes      | `reasoning___[attributename]` |

You MUST include these details explicitly in your reasoning fields:

- **Explicit Evidence**: Quote specific lines or phrases from the document confirming your extraction.
- **Decision Justification**: Clearly justify why specific data was chosen or rejected.
- **Calculations/Transformations**: Document explicitly any computations, unit conversions, or normalizations.
- **Alternative Interpretations**: Explicitly describe any alternative data interpretations considered and why you rejected them.
- **Confidence and Assumptions**: Clearly state your confidence level and explicitly articulate any assumptions.

**Example Reasoning:**

> Found company name 'ACME Corp' explicitly stated in the top-right corner of page 1, matching standard letterhead format. Confirmed by matching signature block ('ACME Corp') at bottom of page 3. Confidence high. Alternative interpretation (e.g., sender's name) explicitly rejected due to explicit labeling 'Client: ACME Corp' on page 1.

---

## Detailed Reasoning Examples

### Array Reasoning (`reasoning___[arrayname]`)

- Explicitly describe how the entire array was identified.
- List explicitly all extracted items with clear details and source references.

**Example:**

```markdown
Identified itemized invoice section clearly demarcated by header "Invoice Items" (page 2, lines 12–17). Extracted items explicitly listed:

1. Office Supplies, quantity 5, unit price $4.99, total $24.95 (line 12)
2. Printer Paper, quantity 1, unit price $5.99, total $5.99 (line 13)
3. Stapler, quantity 1, unit price $4.07, total $4.07 (line 14)

No ambiguity detected.
```

### Array Item Reasoning (`reasoning___item`)

Explicitly document evidence for each individual item:

```markdown
Extracted explicitly from line 12: 'Office Supplies x5 $4.99ea $24.95'. Quantity (5 units) multiplied explicitly by unit price ($4.99) matches listed total ($24.95). Format consistent across invoice, high confidence.
```

---

## Principles for Accurate Extraction

When performing extraction, explicitly follow these core principles:

- **Transparency**: Explicitly document and justify every extraction decision.
- **Precision**: Always verify explicitly using direct quotes from the source document.
- **Conservatism**: Set explicitly fields as `null` when data is explicitly missing or ambiguous—never fabricate or guess.
- **Structure Preservation**: Always maintain explicitly the full schema structure, even when entire nested objects lack data (leaf attributes as null).


## Source Fields

Some leaf fields require you to explicitly provide the source of the data (verbatim from the document).
The idea is to simply provide a verbatim quote from the document, without any additional formatting or commentary, keeping it as close as possible to the original text.
Make sure to reasonably include some surrounding text to provide context about the quote.

You can easily identify the fields that require a source by the `quote___[attributename]` naming pattern.

**Example:**

```json
{
  "quote___name": "NAME:\nJohn Doe",
  "name": "John Doe"
}
```

---

# User Defined System Prompt

'''

    @property
    def user_system_prompt(self) -> str:
        return self.json_schema.get("X-SystemPrompt", "")

    @property
    def schema_system_prompt(self) -> str:
        return (
            self.inference_nlp_data_structure + "\n---\n" + "## Expected output schema as a TypeScript interface for better readability:\n\n" + self.inference_typescript_interface
        )

    @property
    def system_prompt(self) -> str:
        """Returns the system prompt combining custom prompt and TypeScript interface.

        Returns:
            str: The combined system prompt string.
        """
        return self.developer_system_prompt + "\n\n" + self.user_system_prompt + "\n\n" + self.schema_system_prompt

    @property
    def title(self) -> str:
        """Returns the title of the schema.

        Returns:
            str: The schema title or 'NoTitle' if not specified.
        """
        return self.json_schema.get("title", "NoTitle")

    @property
    def _expanded_object_schema(self) -> dict[str, Any]:
        """Returns the schema with all references expanded inline.

        Returns:
            dict[str, Any]: The expanded schema with resolved references. If the schema is not expandable, it is returned as is.
        """
        return expand_refs(copy.deepcopy(self.json_schema))

    @property
    def _reasoning_object_schema(self) -> dict[str, Any]:
        """Returns the schema with inference-specific modifications.

        Returns:
            dict[str, Any]: The modified schema with reasoning fields added to the structure.
        """
        inference_schema = create_reasoning_schema(copy.deepcopy(self._expanded_object_schema))  # Automatically populates the reasoning fields into the structure.
        assert isinstance(inference_schema, dict), "Validation Error: The inference_json_schema is not a dict"
        return inference_schema

    @property
    def _validation_object_schema(self) -> dict[str, Any]:
        """Returns a loose validation schema where all fields are optional.

        This schema ignores all 'required' properties, allowing partial data validation.

        Returns:
            dict[str, Any]: The modified schema for validation purposes.
        """
        # This ignores all 'required' properties (hence making all fields optional)
        # This is a 'loose' validation schema that allows for partial data to be validated.
        _validation_object_schema_ = copy.deepcopy(self._reasoning_object_schema)

        def rec_remove_required(schema: dict[str, Any]) -> None:
            if "required" in schema:
                schema.pop("required")
            if "properties" in schema:
                for prop_schema in schema["properties"].values():
                    rec_remove_required(prop_schema)
            if "items" in schema:
                rec_remove_required(schema["items"])
            if "$defs" in schema:
                for def_schema in schema["$defs"].values():
                    rec_remove_required(def_schema)
            if "anyOf" in schema:
                for anyof_schema in schema["anyOf"]:
                    rec_remove_required(anyof_schema)
            if "allOf" in schema:
                for allof_schema in schema["allOf"]:
                    rec_remove_required(allof_schema)

        rec_remove_required(_validation_object_schema_)
        return _validation_object_schema_

    def _get_pattern_attribute(self, pattern: str, attribute: Literal['X-FieldPrompt', 'X-ReasoningPrompt', 'type']) -> str | None:
        """
        Given a JSON Schema and a pattern (like "my_object.my_array.*.my_property"),
        navigate the schema and return the specified attribute of the identified node.
        """

        # Special case: "*" means the root schema itself
        current_schema = self._expanded_object_schema
        if pattern.strip() == "*":
            if attribute == "X-FieldPrompt":
                return current_schema.get(attribute) or current_schema.get("description")
            return current_schema.get(attribute)

        parts = pattern.split(".")
        index = 0  # Start at the first part

        while index < len(parts):
            part = parts[index]

            if part == "*" or part.isdigit():
                # Handle wildcard case for arrays
                if "items" in current_schema:
                    current_schema = current_schema["items"]
                    index += 1  # Move to the next part
                else:
                    # Invalid use of "*" for the current schema
                    return None
            elif "properties" in current_schema and part in current_schema["properties"]:
                # Handle normal property navigation
                current_schema = current_schema["properties"][part]
                index += 1  # Move to the next part
            else:
                # If we encounter a structure without "properties" or invalid part
                return None

        # At this point, we've navigated to the target node
        if attribute == "X-FieldPrompt":
            return current_schema.get(attribute) or current_schema.get("description")
        elif attribute == "type":
            # Convert schema type to TypeScript type
            return schema_to_ts_type(current_schema, {}, {}, 0, 0, add_field_description=False)
        return current_schema.get(attribute)

    def _set_pattern_attribute(self, pattern: str, attribute: Literal['X-FieldPrompt', 'X-ReasoningPrompt', 'X-SystemPrompt', 'description'], value: str) -> None:
        """Sets an attribute value at a specific path in the schema.

        Args:
            pattern (str): The path pattern to navigate the schema (e.g., "my_object.my_array.*.my_property")
            attribute (Literal): The attribute to set ('description', 'X-FieldPrompt', etc.)
            value (str): The value to set for the attribute
        """
        current_schema = self.json_schema
        definitions = self.json_schema.get("$defs", {})
        parts = pattern.split(".")
        path_stack: list[tuple[str, Any]] = []  # Keep track of how we navigated the schema

        if pattern.strip() == "*":
            # Special case: "*" means the root schema itself
            current_schema[attribute] = value
            return
        assert attribute != "X-SystemPrompt", "Cannot set the X-SystemPrompt attribute other than at the root schema."

        index = 0  # Index for the parts list
        while index < len(parts):
            part = parts[index]
            if part == "*" or part.isdigit():
                # Handle the array case
                if "items" in current_schema:
                    current_schema = current_schema["items"]
                    path_stack.append(("items", None))
                    index += 1  # Move to the next part
                else:
                    return  # Invalid pattern for the current schema

            elif "properties" in current_schema and part in current_schema["properties"]:
                # Handle the properties case
                current_schema = current_schema["properties"][part]
                path_stack.append(("properties", part))
                index += 1  # Move to the next part
            elif "$ref" in current_schema:
                # Handle the $ref case
                ref = current_schema["$ref"]
                assert isinstance(ref, str), "Validation Error: The $ref is not a string"
                assert ref.startswith("#/$defs/"), "Validation Error: The $ref is not a definition reference"
                ref_name = ref.split("/")[-1]
                assert ref_name in definitions, "Validation Error: The $ref is not a definition reference"

                # Count how many times this ref is used in the entire schema
                ref_count = json.dumps(self.json_schema).count(f"\"{ref}\"")

                if ref_count > 1:
                    # Create a unique copy name by appending a number
                    copy_num = 1
                    next_copy_name = f"{ref_name}Copy{copy_num}"
                    while next_copy_name in definitions:
                        copy_num += 1
                        next_copy_name = f"{ref_name}Copy{copy_num}"

                    # Create a copy of the definition
                    def_copy = copy.deepcopy(definitions[ref_name])

                    # Change the title and name of the definition
                    if "title" in def_copy:
                        def_copy["title"] = f"{def_copy['title']} Copy {copy_num}"
                    if "name" in def_copy:
                        def_copy["name"] = next_copy_name

                    # Add the new copy to definitions
                    definitions[next_copy_name] = def_copy

                    # Update the reference
                    current_schema["$ref"] = f"#/$defs/{next_copy_name}"
                    ref_name = next_copy_name
                # Reference is used only once or a copy is created; directly navigate to the definition
                current_schema = definitions[ref_name]
            else:
                # Cannot navigate further; invalid pattern
                return

        # Once we have navigated to the correct node, set the attribute
        current_schema[attribute] = value

    @model_validator(mode="before")
    def validate_schema_and_model(cls, data: Any) -> Any:
        """Validate schema and model logic."""
        # Extract from data
        json_schema: dict[str, Any] | None = data.get('json_schema', None)
        pydantic_model: type[BaseModel] | None = data.get('pydantic_model', None)

        # Check if either json_schema or pydantic_model is provided
        if json_schema and pydantic_model:
            raise ValueError("Cannot provide both json_schema and pydantic_model")

        if not json_schema and not pydantic_model:
            raise ValueError("Must provide either json_schema or pydantic_model")

        if json_schema:
            json_schema = load_json_schema(json_schema)
            data['pydantic_model'] = convert_json_schema_to_basemodel(json_schema)
            data['json_schema'] = json_schema
        if pydantic_model:
            data['pydantic_model'] = pydantic_model
            data['json_schema'] = pydantic_model.model_json_schema()

        return data

    @property
    def messages(self) -> list[ChatCompletionUiformMessage]:
        return [ChatCompletionUiformMessage(role="developer", content=self.system_prompt)]

    @model_validator(mode="after")
    def model_after_validator(self) -> Self:
        # Set the partial_pydantic_model
        self._partial_pydantic_model = convert_basemodel_to_partial_basemodel(self.pydantic_model)

        return self

    def save(self, path: Path | str) -> None:
        """Save a JSON schema to a file.

        Args:
            json_schema: The JSON schema to save, can be a dict, Path, or string
            schema_path: Output path for the schema file
        """
        with open(path, 'w', encoding='utf-8') as f:
            json.dump(self.json_schema, f, ensure_ascii=False, indent=2)



import json
from pathlib import Path
from typing import Any, AsyncGenerator, Generator, TypeVar, Generic, Optional, Union, List, Sequence, cast

from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from openai.types.chat.parsed_chat_completion import ParsedChatCompletionMessage
from openai.types.responses.response import Response
from openai.types.responses.response_input_param import ResponseInputParam, ResponseInputItemParam
from openai.types.responses.response_output_item import ResponseOutputItem
from openai.types.shared_params.response_format_json_schema import ResponseFormatJSONSchema
from pydantic import BaseModel

from ..._resource import AsyncAPIResource, SyncAPIResource
from ..._utils.ai_models import assert_valid_model_extraction
from ..._utils.json_schema import load_json_schema, unflatten_dict
from ..._utils.responses import convert_to_openai_format, convert_from_openai_format, parse_openai_responses_response
from ..._utils.stream_context_managers import as_async_context_manager, as_context_manager
from ...types.chat import ChatCompletionUiformMessage
from ...types.completions import UiChatResponseCreateRequest, UiChatCompletionsRequest
from ...types.documents.extractions import UiParsedChatCompletion, UiParsedChatCompletionChunk, UiParsedChoice, UiResponse
from ...types.standards import PreparedRequest
from ...types.schemas.object import Schema

from typing import Optional, Union
from openai.types.shared_params.reasoning import Reasoning
from openai.types.responses.response_input_param import ResponseInputParam
from openai.types.responses.response_text_config_param import ResponseTextConfigParam
from openai.types.shared_params.response_format_json_schema import ResponseFormatJSONSchema

T = TypeVar('T', bound=BaseModel)

class BaseResponsesMixin:
    def prepare_create(
        self,
        model: str,
        input: Union[str, ResponseInputParam],
        text: ResponseTextConfigParam,
        temperature: float = 0,
        reasoning: Optional[Reasoning] = None,
        n_consensus: int = 1,
        instructions: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> PreparedRequest:
        """
        Prepare a request for the Responses API create method.
        """
        assert_valid_model_extraction(model)

        text_format = text.get("format", None)
        assert text_format is not None, "text.format is required"
        json_schema = text_format.get("schema", None)
        assert json_schema is not None, "text.format.schema is required"

        schema_obj = Schema(json_schema=json_schema)

        if instructions is None:
            instructions = schema_obj.developer_system_prompt
        
        # Create the request object based on the UiChatResponseCreateRequest model
        data = UiChatResponseCreateRequest(
            model=model,
            input=input,
            temperature=temperature,
            stream=False,
            reasoning=reasoning, 
            n_consensus=n_consensus,
            text={
                "format": {
                    "type": "json_schema", 
                    "name": schema_obj.id, 
                    "schema": schema_obj.inference_json_schema, 
                    "strict": True
                }
            },
            instructions=instructions,
        )

        # Validate the request data
        ui_chat_response_create_request = UiChatResponseCreateRequest.model_validate(data)

        return PreparedRequest(
            method="POST", 
            url="/v1/responses", 
            data=ui_chat_response_create_request.model_dump(), 
            idempotency_key=idempotency_key
        )

    def prepare_parse(
        self,
        model: str,
        input: Union[str, ResponseInputParam],
        text_format: type[BaseModel],
        temperature: float = 0,
        reasoning: Optional[Reasoning] = None,
        n_consensus: int = 1,
        instructions: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> PreparedRequest:
        """
        Prepare a request for the Responses API parse method.
        """

        assert_valid_model_extraction(model)

        schema_obj = Schema(pydantic_model=text_format)

        if instructions is None:
            instructions = schema_obj.developer_system_prompt
        
        # Create the request object based on the UiChatResponseCreateRequest model
        data = UiChatResponseCreateRequest(
            model=model,
            input=input,
            temperature=temperature,
            stream=False,
            reasoning=reasoning, 
            n_consensus=n_consensus,
            text={
                "format": {
                    "type": "json_schema", 
                    "name": schema_obj.id, 
                    "schema": schema_obj.inference_json_schema, 
                    "strict": True
                }
            },
            instructions=instructions,
        )

        # Validate the request data
        ui_chat_response_create_request = UiChatResponseCreateRequest.model_validate(data)

        return PreparedRequest(
            method="POST", 
            url="/v1/responses", 
            data=ui_chat_response_create_request.model_dump(), 
            idempotency_key=idempotency_key
        )
        

        return PreparedRequest(
            method="POST", 
            url="/v1/completions", 
            data=ui_chat_completions_request.model_dump(), 
            idempotency_key=idempotency_key
        )


class Responses(SyncAPIResource, BaseResponsesMixin):
    """UiForm Responses API compatible with OpenAI Responses API"""

    def create(
        self,
        model: str,
        input: Union[str, ResponseInputParam],
        text: ResponseTextConfigParam,
        temperature: float = 0,
        reasoning: Optional[Reasoning] = None,
        n_consensus: int = 1,
        instructions: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> Response:
        """
        Create a completion using the UiForm API with OpenAI Responses API compatible interface.
        
        Args:
            model: The model to use
            input: The input text or message array
            temperature: Model temperature setting (0-1)
            reasoning: The effort level for the model to reason about the input data
            n_consensus: Number of consensus models to use
            text: The response format configuration
            instructions: Optional system instructions
            idempotency_key: Idempotency key for request
        Returns:
            Response: OpenAI Responses API compatible response
        """
        request = self.prepare_create(
            model=model,
            input=input,
            temperature=temperature,
            reasoning=reasoning,
            text=text,
            instructions=instructions,
            n_consensus=n_consensus,
            idempotency_key=idempotency_key,
        )

        result = self._client._prepared_request(request)
        response = UiResponse.model_validate(result)
        
        return response
            
    def parse(
        self,
        model: str,
        input: Union[str, ResponseInputParam],
        text_format: type[T],
        temperature: float = 0,
        reasoning: Optional[Reasoning] = None,
        n_consensus: int = 1,
        instructions: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> Response:
        """
        Parse content using the UiForm API with OpenAI Responses API compatible interface.
        
        Args:
            model: The model to use
            input: The input text or message array
            text_format: The Pydantic model defining the expected output format
            temperature: Model temperature setting (0-1)
            reasoning_effort: The effort level for the model to reason about the input data
            n_consensus: Number of consensus models to use
            instructions: Optional system instructions
            idempotency_key: Idempotency key for request
            
        Returns:
            Response: OpenAI Responses API compatible response with parsed content
        """
        request = self.prepare_parse(
            model=model,
            input=input,
            temperature=temperature,
            reasoning=reasoning,
            text_format=text_format,
            instructions=instructions,
            n_consensus=n_consensus,
            idempotency_key=idempotency_key,
        )

        result = self._client._prepared_request(request)
        response = UiResponse.model_validate(result)
        
        return response


class AsyncResponses(AsyncAPIResource, BaseResponsesMixin):
    """UiForm Responses API compatible with OpenAI Responses API for async usage"""

    async def create(
        self,
        model: str,
        input: Union[str, ResponseInputParam],
        text: ResponseTextConfigParam,
        temperature: float = 0,
        reasoning: Optional[Reasoning] = None,
        n_consensus: int = 1,
        instructions: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> UiResponse:
        """
        Create a completion using the UiForm API asynchronously with OpenAI Responses API compatible interface.
        
        Args:
            model: The model to use
            input: The input text or message array
            text: The response format configuration
            temperature: Model temperature setting (0-1)
            reasoning: The effort level for the model to reason about the input data
            n_consensus: Number of consensus models to use
            instructions: Optional system instructions
            idempotency_key: Idempotency key for request
            
        Returns:
            Response: OpenAI Responses API compatible response
        """
        request = self.prepare_create(
            model=model,
            input=input,
            temperature=temperature,
            reasoning=reasoning,
            text=text,
            instructions=instructions,
            n_consensus=n_consensus,
            idempotency_key=idempotency_key,
        )

        result = await self._client._prepared_request(request)
        response = UiResponse.model_validate(result)
        return response
        

            
    async def parse(
        self,
        model: str,
        input: Union[str, ResponseInputParam],
        text_format: type[BaseModel],
        temperature: float = 0,
        reasoning: Optional[Reasoning] = None,
        n_consensus: int = 1,
        instructions: Optional[str] = None,
        idempotency_key: Optional[str] = None,
    ) -> UiResponse:
        """
        Parse content using the UiForm API asynchronously with OpenAI Responses API compatible interface.
        
        Args:
            model: The model to use
            input: The input text or message array
            text_format: The Pydantic model defining the expected output format
            temperature: Model temperature setting (0-1)
            reasoning: The effort level for the model to reason about the input data
            n_consensus: Number of consensus models to use
            instructions: Optional system instructions
            idempotency_key: Idempotency key for request
            
        Returns:
            Response: OpenAI Responses API compatible response with parsed content
        """
        request = self.prepare_parse(
            model=model,
            input=input,
            temperature=temperature,
            reasoning=reasoning,
            text_format=text_format,
            instructions=instructions,
            n_consensus=n_consensus,
            idempotency_key=idempotency_key,
        )

        result = await self._client._prepared_request(request)
        response = UiResponse.model_validate(result)
        return response
        


from typing import Literal, Optional

from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from pydantic import BaseModel, Field

from ..modalities import Modality


class InferenceSettings(BaseModel):
    model: str = "gpt-4o-mini"
    temperature: float = 0.0
    modality: Modality
    image_resolution_dpi: int = 96
    browser_canvas: Literal['A3', 'A4', 'A5'] = 'A4'
    reasoning_effort: ChatCompletionReasoningEffort = "medium"


class AnnotationInputData(BaseModel):
    dataset_id: str
    files_ids: Optional[list[str]] = None
    upsert: bool = False
    inference_settings: InferenceSettings



import base64
import datetime
from io import IOBase
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

import httpx
import PIL.Image
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from pydantic import HttpUrl

from ..._resource import AsyncAPIResource, SyncAPIResource
from ..._utils.ai_models import assert_valid_model_extraction
from ..._utils.mime import MIMEData, prepare_mime_document
from ...types.logs import ProcessorConfig, UpdateProcessorRequest
from ...types.pagination import ListMetadata
from ...types.documents.extractions import UiParsedChatCompletion, DocumentExtractRequest
from pydantic import BaseModel

# from ...types.documents.extractions import DocumentExtractResponse
from ...types.mime import BaseMIMEData
from ...types.modalities import Modality
from ...types.standards import PreparedRequest
from .automations import Automations, AsyncAutomations


class ListProcessors(BaseModel):
    data: List[ProcessorConfig]
    list_metadata: ListMetadata


class ProcessorsMixin:
    def prepare_create(
        self,
        name: str,
        json_schema: Dict[str, Any],
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        image_resolution_dpi: Optional[int] = 96,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = 'A4',
        n_consensus: int = 1,
    ) -> PreparedRequest:
        assert_valid_model_extraction(model)

        data = {
            "name": name,
            "json_schema": json_schema,
            "modality": modality,
            "model": model,
            "temperature": temperature,
            "reasoning_effort": reasoning_effort,
            "image_resolution_dpi": image_resolution_dpi,
            "browser_canvas": browser_canvas,
            "n_consensus": n_consensus,
        }

        request = ProcessorConfig.model_validate(data)
        return PreparedRequest(method="POST", url="/v1/processors", data=request.model_dump(mode='json'))

    def prepare_list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        # Filtering parameters
        name: Optional[str] = None,
        modality: Optional[str] = None,
        model: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> PreparedRequest:
        params = {
            "before": before,
            "after": after,
            "limit": limit,
            "order": order,
            "name": name,
            "modality": modality,
            "model": model,
            "schema_id": schema_id,
            "schema_data_id": schema_data_id,
        }
        # Remove None values
        params = {k: v for k, v in params.items() if v is not None}

        return PreparedRequest(method="GET", url="/v1/processors", params=params)

    def prepare_get(self, processor_id: str) -> PreparedRequest:
        """Get a specific processor configuration.

        Args:
            processor_id: ID of the processor

        Returns:
            ProcessorConfig: The processor configuration
        """
        return PreparedRequest(method="GET", url=f"/v1/processors/{processor_id}")

    def prepare_update(
        self,
        processor_id: str,
        name: Optional[str] = None,
        modality: Optional[Modality] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        model: Optional[str] = None,
        json_schema: Optional[Dict[str, Any]] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
        n_consensus: Optional[int] = None,
    ) -> PreparedRequest:
        data: dict[str, Any] = {}

        if name is not None:
            data["name"] = name
        if modality is not None:
            data["modality"] = modality
        if image_resolution_dpi is not None:
            data["image_resolution_dpi"] = image_resolution_dpi
        if browser_canvas is not None:
            data["browser_canvas"] = browser_canvas
        if model is not None:
            assert_valid_model_extraction(model)
            data["model"] = model
        if json_schema is not None:
            data["json_schema"] = json_schema
        if temperature is not None:
            data["temperature"] = temperature
        if reasoning_effort is not None:
            data["reasoning_effort"] = reasoning_effort
        if n_consensus is not None:
            data["n_consensus"] = n_consensus
        request = UpdateProcessorRequest.model_validate(data)
        return PreparedRequest(method="PUT", url=f"/v1/processors/{processor_id}", data=request.model_dump(mode='json'))

    def prepare_delete(self, processor_id: str) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"/v1/processors/{processor_id}")

    def prepare_submit(
        self,
        processor_id: str,
        document: Optional[Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl] = None,
        documents: Optional[List[Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl]] = None,
        temperature: Optional[float] = None,
        seed: Optional[int] = None,
        store: bool = True,
    ) -> PreparedRequest:
        """Prepare a request to submit documents to a processor.

        Args:
            processor_id: ID of the processor
            document: Single document to process (mutually exclusive with documents)
            documents: List of documents to process (mutually exclusive with document)
            temperature: Optional temperature override
            seed: Optional seed for reproducibility
            store: Whether to store the results

        Returns:
            PreparedRequest: The prepared request
        """
        # Validate that either document or documents is provided, but not both
        if not document and not documents:
            raise ValueError("Either 'document' or 'documents' must be provided")
        
        if document and documents:
            raise ValueError("Provide either 'document' (single) or 'documents' (multiple), not both")
        
        # Prepare form data parameters
        form_data = {
            "temperature": temperature,
            "seed": seed,
            "store": store,
        }
        # Remove None values
        form_data = {k: v for k, v in form_data.items() if v is not None}

        # Prepare files for upload
        files = {}
        if document:
            # Convert document to MIMEData if needed
            mime_document = prepare_mime_document(document)
            # Single document upload
            files["document"] = (
                mime_document.filename,
                base64.b64decode(mime_document.content),
                mime_document.mime_type
            )
        elif documents:
            # Multiple documents upload - httpx supports multiple files with same field name using a list
            files_list = []
            for doc in documents:
                # Convert each document to MIMEData if needed
                mime_doc = prepare_mime_document(doc)
                files_list.append((
                    "documents",  # field name
                    (
                        mime_doc.filename,
                        base64.b64decode(mime_doc.content),
                        mime_doc.mime_type
                    )
                ))
            files = files_list

        url = f"/v1/processors/{processor_id}/submit"
        # if stream:
        #     url = f"/v1/processors/{processor_id}/submit/stream"
        
        return PreparedRequest(method="POST", url=url, form_data=form_data, files=files)


class Processors(SyncAPIResource, ProcessorsMixin):
    """Processors API wrapper for managing processor configurations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.automations = Automations(client=client)

    def create(
        self,
        name: str,
        json_schema: Dict[str, Any],
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        image_resolution_dpi: Optional[int] = 96,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = 'A4',
        n_consensus: int = 1,
    ) -> ProcessorConfig:
        """Create a new processor configuration.

        Args:
            name: Name of the processor
            json_schema: JSON schema for the processor
            image_resolution_dpi: Optional image resolution DPI
            browser_canvas: Optional browser canvas size
            modality: Processing modality (currently only "native" supported)
            model: AI model to use for processing
            temperature: Model temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.
            n_consensus: Number of consensus required to validate the data
        Returns:
            ProcessorConfig: The created processor configuration
        """
        request = self.prepare_create(name, json_schema, modality, model, temperature, reasoning_effort, image_resolution_dpi, browser_canvas, n_consensus)
        response = self._client._prepared_request(request)
        print(f"Processor ID: {response['id']}. Processor available at https://www.uiform.com/dashboard/processors/{response['id']}")
        return ProcessorConfig.model_validate(response)

    def list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        name: Optional[str] = None,
        modality: Optional[str] = None,
        model: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> ListProcessors:
        """List processor configurations with pagination support.

        Args:
            before: Optional cursor for pagination before a specific processor ID
            after: Optional cursor for pagination after a specific processor ID
            limit: Optional limit on number of results (max 100)
            order: Optional sort order ("asc" or "desc")
            name: Optional filter by processor name
            modality: Optional filter by modality
            model: Optional filter by model
            schema_id: Optional filter by schema ID
            schema_data_id: Optional filter by schema data ID

        Returns:
            ListProcessors: Paginated list of processor configurations with metadata
        """
        request = self.prepare_list(before, after, limit, order, name, modality, model, schema_id, schema_data_id)
        response = self._client._prepared_request(request)
        return ListProcessors.model_validate(response)

    def get(self, processor_id: str) -> ProcessorConfig:
        """Get a specific processor configuration.

        Args:
            processor_id: ID of the processor

        Returns:
            ProcessorConfig: The processor configuration
        """
        request = self.prepare_get(processor_id)
        response = self._client._prepared_request(request)
        return ProcessorConfig.model_validate(response)

    def update(
        self,
        processor_id: str,
        name: Optional[str] = None,
        modality: Optional[Modality] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        model: Optional[str] = None,
        json_schema: Optional[Dict[str, Any]] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
        n_consensus: Optional[int] = None,
    ) -> ProcessorConfig:
        """Update a processor configuration.

        Args:
            processor_id: ID of the processor to update
            name: New name for the processor
            modality: New processing modality
            image_resolution_dpi: New image resolution DPI
            browser_canvas: New browser canvas size
            model: New AI model
            json_schema: New JSON schema for the processor
            temperature: New temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.
            n_consensus: New number of consensus required
        Returns:
            ProcessorConfig: The updated processor configuration
        """
        request = self.prepare_update(processor_id, name, modality, image_resolution_dpi, browser_canvas, model, json_schema, temperature, reasoning_effort, n_consensus)
        response = self._client._prepared_request(request)
        return ProcessorConfig.model_validate(response)

    def delete(self, processor_id: str) -> None:
        """Delete a processor configuration.

        Args:
            processor_id: ID of the processor to delete
        """
        request = self.prepare_delete(processor_id)
        self._client._prepared_request(request)
        print(f"Processor Deleted. ID: {processor_id}")

    def submit(
        self,
        processor_id: str,
        document: Optional[Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl] = None,
        documents: Optional[List[Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl]] = None,
        temperature: Optional[float] = None,
        seed: Optional[int] = None,
        store: bool = True,
    ) -> UiParsedChatCompletion:
        """Submit documents to a processor for processing.

        Args:
            processor_id: ID of the processor
            document: Single document to process (mutually exclusive with documents)
            documents: List of documents to process (mutually exclusive with document)
            temperature: Optional temperature override
            seed: Optional seed for reproducibility
            store: Whether to store the results

        Returns:
            UiParsedChatCompletion: The processing result
        """
        request = self.prepare_submit(
            processor_id=processor_id,
            document=document,
            documents=documents,
            temperature=temperature,
            seed=seed,
            store=store
        )
        response = self._client._prepared_request(request)
        return UiParsedChatCompletion.model_validate(response)


class AsyncProcessors(AsyncAPIResource, ProcessorsMixin):
    """Async Processors API wrapper for managing processor configurations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.automations = AsyncAutomations(client=client)

    async def create(
        self,
        name: str,
        json_schema: Dict[str, Any],
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        image_resolution_dpi: Optional[int] = 96,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = 'A4',
        n_consensus: int = 1,
    ) -> ProcessorConfig:
        request = self.prepare_create(name, json_schema, modality, model, temperature, reasoning_effort, image_resolution_dpi, browser_canvas, n_consensus)
        response = await self._client._prepared_request(request)
        print(f"Processor ID: {response['id']}. Processor available at https://www.uiform.com/dashboard/processors/{response['id']}")
        
        return ProcessorConfig.model_validate(response)

    async def list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        name: Optional[str] = None,
        modality: Optional[str] = None,
        model: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> ListProcessors:
        request = self.prepare_list(before, after, limit, order, name, modality, model, schema_id, schema_data_id)
        response = await self._client._prepared_request(request)
        return ListProcessors.model_validate(response)

    async def get(self, processor_id: str) -> ProcessorConfig:
        request = self.prepare_get(processor_id)
        response = await self._client._prepared_request(request)
        return ProcessorConfig.model_validate(response)

    async def update(
        self,
        processor_id: str,
        name: Optional[str] = None,
        modality: Optional[Modality] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        model: Optional[str] = None,
        json_schema: Optional[Dict[str, Any]] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
        n_consensus: Optional[int] = None,
    ) -> ProcessorConfig:
        request = self.prepare_update(processor_id, name, modality, image_resolution_dpi, browser_canvas, model, json_schema, temperature, reasoning_effort, n_consensus)
        response = await self._client._prepared_request(request)
        return ProcessorConfig.model_validate(response)

    async def delete(self, processor_id: str) -> None:
        request = self.prepare_delete(processor_id)
        await self._client._prepared_request(request)
        print(f"Processor Deleted. ID: {processor_id}")

    async def submit(
        self,
        processor_id: str,
        document: Optional[Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl] = None,
        documents: Optional[List[Path | str | bytes | IOBase | MIMEData | PIL.Image.Image | HttpUrl]] = None,
        temperature: Optional[float] = None,
        seed: Optional[int] = None,
        store: bool = True,
    ) -> UiParsedChatCompletion:
        """Submit documents to a processor for processing.

        Args:
            processor_id: ID of the processor
            document: Single document to process (mutually exclusive with documents)
            documents: List of documents to process (mutually exclusive with document)
            temperature: Optional temperature override
            seed: Optional seed for reproducibility
            store: Whether to store the results

        Returns:
            UiParsedChatCompletion: The processing result
        """
        request = self.prepare_submit(
            processor_id=processor_id,
            document=document,
            documents=documents,
            temperature=temperature,
            seed=seed,
            store=store
        )
        response = await self._client._prepared_request(request)
        return UiParsedChatCompletion.model_validate(response)



from typing import Any

import nanoid  # type: ignore
import pytest
from pydantic import HttpUrl

from uiform import UiForm


@pytest.mark.asyncio
async def test_mailboxes_crud(sync_client: UiForm, company_json_schema: dict[str, Any], booking_confirmation_file_path: str) -> None:
    test_idx = nanoid.generate().lower()
    email_address = f"bert2_{test_idx}@devmail.uiform.com"
    webhook_url = HttpUrl('http://localhost:4000/product')

    # Create
    mailbox = sync_client.processors.automations.mailboxes.create(email=email_address, json_schema=company_json_schema, webhook_url=webhook_url)
    try:
        assert mailbox.email == email_address
        assert mailbox.webhook_url == webhook_url
        # Read
        mailbox = sync_client.processors.automations.mailboxes.get(email_address)
        assert mailbox.email == email_address
        # Update
        mailbox = sync_client.processors.automations.mailboxes.update(email_address, webhook_url=HttpUrl('http://localhost:4000/product2'))
        assert mailbox.webhook_url == HttpUrl('http://localhost:4000/product2')

        # TODO: send and email to email_address (need sendgrid account)

        # Something like we did in test_automations_links.py, but not quite:
        # with open(booking_confirmation_file_path, "rb") as f:
        #     async with httpx.AsyncClient(timeout=240) as client:
        #         usr_pwd_enc = base64.b64encode(f"{name}:password".encode("utf-8")).decode("utf-8")
        #         headers = {
        #             "Authorization": f"Basic {usr_pwd_enc}",
        #         }
        #         files = {
        #             "file": f
        #         }
        #         response = await client.post(
        #             sync_client.base_url + f"/v1/processors/automations/links/parse/{link_id}",
        #             files=files,
        #             headers=headers
        #         )
        #         assert response.status_code == 200

        # Delete
        sync_client.processors.automations.mailboxes.delete(email_address)
        with pytest.raises(Exception):
            sync_client.processors.automations.mailboxes.get(email_address)
    finally:
        # Delete the mailbox
        try:
            sync_client.processors.automations.mailboxes.delete(email_address)
        except Exception:
            pass



import json
from io import IOBase
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from PIL.Image import Image
from pydantic import HttpUrl

from ...._resource import AsyncAPIResource, SyncAPIResource
from ...._utils.ai_models import assert_valid_model_extraction
from ....types.automations.outlook import FetchParams, ListOutlooks, MatchParams, Outlook, UpdateOutlookRequest
from ....types.logs import AutomationLog
from ....types.modalities import Modality
from ....types.standards import PreparedRequest


class OutlooksMixin:
    def prepare_create(
        self,
        name: str,
        json_schema: Dict[str, Any],
        webhook_url: HttpUrl,
        # email specific opitonals Fields
        authorized_domains: List[str] = [],
        authorized_emails: List[str] = [],
        # HTTP Config Optional Fields
        webhook_headers: Dict[str, str] = {},
        # DocumentExtraction Config
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        # Optional Fields for data integration
        match_params: Optional[List[MatchParams]] = None,
        fetch_params: Optional[List[FetchParams]] = None,
        layout_schema: Optional[Dict[str, Any]] = None,
    ) -> PreparedRequest:
        assert_valid_model_extraction(model)

        data = {
            "name": name,
            "webhook_url": webhook_url,
            "webhook_headers": webhook_headers,
            "json_schema": json_schema,
            "authorized_domains": authorized_domains,
            "authorized_emails": authorized_emails,
            "image_resolution_dpi": image_resolution_dpi,
            "browser_canvas": browser_canvas,
            "modality": modality,
            "model": model,
            "temperature": temperature,
            "reasoning_effort": reasoning_effort,
            "layout_schema": layout_schema,
        }

        if match_params is not None:
            data["match_params"] = match_params
        if fetch_params is not None:
            data["fetch_params"] = fetch_params

        # Validate the data
        outlook_data = Outlook.model_validate(data)
        return PreparedRequest(method="POST", url="/v1/deployments/outlook", data=outlook_data.model_dump(mode="json"))

    def prepare_list(
        self,
        before: str | None = None,
        after: str | None = None,
        limit: int = 10,
        order: Literal["asc", "desc"] | None = "desc",
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> PreparedRequest:
        params = {
            "before": before,
            "after": after,
            "limit": limit,
            "order": order,
            "name": name,
            "webhook_url": webhook_url,
            "schema_id": schema_id,
            "schema_data_id": schema_data_id,
        }
        # Remove None values
        params = {k: v for k, v in params.items() if v is not None}

        return PreparedRequest(method="GET", url="/v1/deployments/outlook", params=params)

    def prepare_get(self, outlook_id: str) -> PreparedRequest:
        return PreparedRequest(method="GET", url=f"/v1/processors/automations/outlook/{outlook_id}")

    def prepare_update(
        self,
        outlook_id: str,
        name: Optional[str] = None,
        webhook_url: Optional[HttpUrl] = None,
        webhook_headers: Optional[Dict[str, str]] = None,
        authorized_domains: Optional[List[str]] = None,
        authorized_emails: Optional[List[str]] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Optional[Modality] = None,
        model: Optional[str] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
        json_schema: Optional[Dict[str, Any]] = None,
        match_params: Optional[List[MatchParams]] = None,
        fetch_params: Optional[List[FetchParams]] = None,
        layout_schema: Optional[Dict[str, Any]] = None,
    ) -> PreparedRequest:
        data: dict[str, Any] = {}
        if name is not None:
            data["name"] = name
        if webhook_url is not None:
            data["webhook_url"] = webhook_url
        if webhook_headers is not None:
            data["webhook_headers"] = webhook_headers
        if authorized_domains is not None:
            data["authorized_domains"] = authorized_domains
        if authorized_emails is not None:
            data["authorized_emails"] = authorized_emails
        if image_resolution_dpi is not None:
            data["image_resolution_dpi"] = image_resolution_dpi
        if browser_canvas is not None:
            data["browser_canvas"] = browser_canvas
        if modality is not None:
            data["modality"] = modality
        if model is not None:
            assert_valid_model_extraction(model)
            data["model"] = model
        if temperature is not None:
            data["temperature"] = temperature
        if json_schema is not None:
            data["json_schema"] = json_schema
        if match_params is not None:
            data["match_params"] = match_params
        if fetch_params is not None:
            data["fetch_params"] = fetch_params
        if reasoning_effort is not None:
            data["reasoning_effort"] = reasoning_effort
        if layout_schema is not None:
            data["layout_schema"] = layout_schema

        update_outlook_request = UpdateOutlookRequest.model_validate(data)

        return PreparedRequest(method="PUT", url=f"/v1/processors/automations/outlook/{outlook_id}", data=update_outlook_request.model_dump(mode="json"))

    def prepare_delete(self, outlook_id: str) -> PreparedRequest:
        return PreparedRequest(method="DELETE", url=f"/v1/processors/automations/outlook/{outlook_id}")

class Outlooks(SyncAPIResource, OutlooksMixin):
    """Outlook API wrapper for managing outlook automation configurations"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)

    def create(
        self,
        name: str,
        json_schema: Dict[str, Any],
        webhook_url: HttpUrl,
        authorized_domains: List[str] = [],
        authorized_emails: List[str] = [],
        webhook_headers: Dict[str, str] = {},
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        match_params: Optional[List[MatchParams]] = None,
        fetch_params: Optional[List[FetchParams]] = None,
    ) -> Outlook:
        """Create a new outlook automation configuration.

        Args:
            name: Name of the outlook plugin
            json_schema: JSON schema to validate extracted data
            webhook_url: Webhook URL to receive processed data
            webhook_headers: Webhook headers to send with processed data
            authorized_domains: List of authorized domains
            authorized_emails: List of authorized emails
            image_resolution_dpi: Optional image resolution DPI
            browser_canvas: Optional browser canvas size
            modality: Processing modality (currently only "native" supported)
            model: AI model to use for processing
            temperature: Model temperature setting
            reasoning_effort: The effort level for the model to reason about the input data.
            match_params: List of match parameters for the outlook automation
            fetch_params: List of fetch parameters for the outlook automation
        Returns:
            Outlook: The created outlook plugin configuration
        """

        request = self.prepare_create(
            name,
            json_schema,
            webhook_url,
            authorized_domains,
            authorized_emails,
            webhook_headers,
            image_resolution_dpi,
            browser_canvas,
            modality,
            model,
            temperature,
            reasoning_effort,
            match_params,
            fetch_params,
        )
        response = self._client._prepared_request(request)

        print(f"Outlook automation created. Outlook available at https://www.uiform.com/dashboard/processors/{response['id']}")

        return Outlook.model_validate(response)

    def list(
        self,
        before: str | None = None,
        after: str | None = None,
        limit: int = 10,
        order: Literal["asc", "desc"] | None = "desc",
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> ListOutlooks:
        """List all outlook automation configurations.

        Args:
            before: Optional cursor for pagination - get results before this log ID
            after: Optional cursor for pagination - get results after this log ID
            limit: Maximum number of logs to return (1-100, default 10)
            order: Sort order by creation time - "asc" or "desc" (default "desc")
            name: Optional name filter
            webhook_url: Optional webhook URL filter
            schema_id: Optional schema ID filter
            schema_data_id: Optional schema data ID filter
            match_params: Optional list of match parameters for the outlook automation
            fetch_params: Optional list of fetch parameters for the outlook automation
        Returns:
            List[Outlook]: List of outlook plugin configurations
        """
        request = self.prepare_list(before, after, limit, order, name, webhook_url, schema_id, schema_data_id)
        response = self._client._prepared_request(request)
        return ListOutlooks.model_validate(response)

    def get(self, id: str) -> Outlook:
        """Get a specific outlook automation configuration.

        Args:
            id: ID of the outlook plugin

        Returns:
            Outlook: The outlook plugin configuration
        """
        request = self.prepare_get(id)
        response = self._client._prepared_request(request)
        return Outlook.model_validate(response)

    def update(
        self,
        outlook_id: str,
        name: Optional[str] = None,
        webhook_url: Optional[HttpUrl] = None,
        webhook_headers: Optional[Dict[str, str]] = None,
        authorized_domains: Optional[List[str]] = None,
        authorized_emails: Optional[List[str]] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Optional[Modality] = None,
        model: Optional[str] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
        json_schema: Optional[Dict[str, Any]] = None,
        match_params: Optional[List[MatchParams]] = None,
        fetch_params: Optional[List[FetchParams]] = None,
        layout_schema: Optional[Dict[str, Any]] = None,
    ) -> Outlook:
        """Update an outlook automation configuration.

        Args:
            outlook_id: ID of the outlook plugin to update
            name: New name for the outlook plugin
            webhook_url: New webhook URL
            webhook_headers: New webhook headers
            authorized_domains: New authorized domains
            authorized_emails: New authorized emails
            image_resolution_dpi: New image resolution DPI
            browser_canvas: New browser canvas size
            modality: New processing modality
            model: New AI model
            temperature: New temperature setting
            reasoning_effort: New reasoning effort
            json_schema: New JSON schema
            match_params: New match parameters for the outlook automation
            fetch_params: New fetch parameters for the outlook automation
            layout_schema: New layout schema for the outlook automation

        Returns:
            Outlook: The updated outlook plugin configuration
        """
        request = self.prepare_update(
            outlook_id,
            name,
            webhook_url,
            webhook_headers,
            authorized_domains,
            authorized_emails,
            image_resolution_dpi,
            browser_canvas,
            modality,
            model,
            temperature,
            reasoning_effort,
            json_schema,
            match_params,
            fetch_params,
            layout_schema,
        )
        response = self._client._prepared_request(request)
        return Outlook.model_validate(response)

    def delete(self, outlook_id: str) -> None:
        """Delete an outlook automation configuration.

        Args:
            outlook_id: ID of the outlook plugin to delete
        """
        request = self.prepare_delete(outlook_id)
        response = self._client._prepared_request(request)
        return None


class AsyncOutlooks(AsyncAPIResource, OutlooksMixin):
    def __init__(self, client: Any) -> None:
        super().__init__(client=client)

    async def create(
        self,
        name: str,
        json_schema: Dict[str, Any],
        webhook_url: HttpUrl,
        authorized_domains: List[str] = [],
        authorized_emails: List[str] = [],
        webhook_headers: Dict[str, str] = {},
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Modality = "native",
        model: str = "gpt-4o-mini",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        match_params: List[MatchParams] = [],
        fetch_params: List[FetchParams] = [],
    ) -> Outlook:
        request = self.prepare_create(
            name,
            json_schema,
            webhook_url,
            authorized_domains,
            authorized_emails,
            webhook_headers,
            image_resolution_dpi,
            browser_canvas,
            modality,
            model,
            temperature,
            reasoning_effort,
            match_params,
            fetch_params,
        )
        response = await self._client._prepared_request(request)
        print(f"Outlook automation created. Outlook available at https://www.uiform.com/dashboard/processors/{response['id']}")
        return Outlook.model_validate(response)

    async def list(
        self,
        before: str | None = None,
        after: str | None = None,
        limit: int = 10,
        order: Literal["asc", "desc"] | None = "desc",
        name: Optional[str] = None,
        webhook_url: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> ListOutlooks:
        request = self.prepare_list(before, after, limit, order, name, webhook_url, schema_id, schema_data_id)
        response = await self._client._prepared_request(request)
        return ListOutlooks.model_validate(response)

    async def get(self, outlook_id: str) -> Outlook:
        request = self.prepare_get(outlook_id)
        response = await self._client._prepared_request(request)
        return Outlook.model_validate(response)

    async def update(
        self,
        outlook_id: str,
        name: Optional[str] = None,
        webhook_url: Optional[HttpUrl] = None,
        webhook_headers: Optional[Dict[str, str]] = None,
        authorized_domains: Optional[List[str]] = None,
        authorized_emails: Optional[List[str]] = None,
        image_resolution_dpi: Optional[int] = None,
        browser_canvas: Optional[Literal['A3', 'A4', 'A5']] = None,
        modality: Optional[Modality] = None,
        model: Optional[str] = None,
        temperature: Optional[float] = None,
        reasoning_effort: Optional[ChatCompletionReasoningEffort] = None,
        json_schema: Optional[Dict[str, Any]] = None,
        match_params: Optional[List[MatchParams]] = None,
        fetch_params: Optional[List[FetchParams]] = None,
        layout_schema: Optional[Dict[str, Any]] = None,
    ) -> Outlook:
        request = self.prepare_update(
            outlook_id,
            name,
            webhook_url,
            webhook_headers,
            authorized_domains,
            authorized_emails,
            image_resolution_dpi,
            browser_canvas,
            modality,
            model,
            temperature,
            reasoning_effort,
            json_schema,
            match_params,
            fetch_params,
            layout_schema,
        )
        response = await self._client._prepared_request(request)
        return Outlook.model_validate(response)

    async def delete(self, outlook_id: str) -> None:
        request = self.prepare_delete(outlook_id)
        await self._client._prepared_request(request)
        return None



from typing import Literal, Optional

import nanoid  # type: ignore
from pydantic import BaseModel, Field

from ..logs import AutomationConfig, UpdateAutomationRequest
from ..pagination import ListMetadata


class Link(AutomationConfig):
    object: Literal['automation.link'] = "automation.link"
    id: str = Field(default_factory=lambda: "lnk_" + nanoid.generate(), description="Unique identifier for the extraction link")

    # Link Specific Config
    password: Optional[str] = Field(None, description="Password to access the link")


class ListLinks(BaseModel):
    data: list[Link]
    list_metadata: ListMetadata


# Inherits from the methods of UpdateAutomationRequest
class UpdateLinkRequest(UpdateAutomationRequest):
    # ------------------------------
    # Link Config
    # ------------------------------
    password: Optional[str] = None



from io import IOBase
from pathlib import Path
from typing import Any, List, Optional, Sequence

import PIL.Image
from pydantic import BaseModel

from .._resource import AsyncAPIResource, SyncAPIResource
from .._utils.ai_models import assert_valid_model_schema_generation
from .._utils.json_schema import load_json_schema
from .._utils.mime import prepare_mime_document_list
from ..types.modalities import Modality
from ..types.mime import MIMEData
from ..types.schemas.generate import GenerateSchemaRequest
from ..types.schemas.enhance import EnhanceSchemaConfig, EnhanceSchemaConfigDict, EnhanceSchemaRequest
from ..types.schemas.evaluate import EvaluateSchemaRequest, EvaluateSchemaResponse
from ..types.schemas.object import Schema
from ..types.standards import PreparedRequest
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort


class SchemasMixin:
    def prepare_generate(
        self,
        documents: Sequence[Path | str | bytes | MIMEData | IOBase | PIL.Image.Image],
        instructions: str | None = None,
        model: str = "gpt-4o-2024-11-20",
        temperature: float = 0,
        modality: Modality = "native",
    ) -> PreparedRequest:
        assert_valid_model_schema_generation(model)
        mime_documents = prepare_mime_document_list(documents)
        data = {
            "documents": [doc.model_dump() for doc in mime_documents],
            "instructions": instructions if instructions else None,
            "model": model,
            "temperature": temperature,
            "modality": modality,
        }
        GenerateSchemaRequest.model_validate(data)
        return PreparedRequest(method="POST", url="/v1/schemas/generate", data=data)

    def prepare_evaluate(
        self,
        documents: Sequence[Path | str | bytes | MIMEData | IOBase | PIL.Image.Image],
        json_schema: dict[str, Any],
        ground_truths: list[dict[str, Any]] | None = None,
        model: str = "gpt-4o-mini",
        temperature: float = 0.0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        modality: Modality = "native",
        image_resolution_dpi: int = 96,
        browser_canvas: str = "A4",
        n_consensus: int = 1,
    ) -> PreparedRequest:
        # Assert that if ground_truths is not None, it has the same length as documents
        if ground_truths is not None and len(documents) != len(ground_truths):
            raise ValueError("Number of documents must match number of ground truths")

        mime_documents = prepare_mime_document_list(documents)
        data = {
            "documents": [doc.model_dump() for doc in mime_documents],
            "ground_truths": ground_truths,
            "model": model,
            "temperature": temperature,
            "reasoning_effort": reasoning_effort,
            "modality": modality,
            "image_resolution_dpi": image_resolution_dpi,
            "browser_canvas": browser_canvas,
            "n_consensus": n_consensus,
            "json_schema": json_schema,
        }
        EvaluateSchemaRequest.model_validate(data)
        return PreparedRequest(method="POST", url="/v1/schemas/evaluate", data=data)

    def prepare_enhance(
        self,
        json_schema: dict[str, Any] | Path | str,
        documents: Sequence[Path | str | bytes | MIMEData | IOBase | PIL.Image.Image],
        ground_truths: list[dict[str, Any]] | None,
        instructions: str | None,
        model: str,
        temperature: float,
        modality: Modality,
        flat_likelihoods: list[dict[str, float]] | dict[str, float] | None,
        tools_config: EnhanceSchemaConfig,
    ) -> PreparedRequest:
        assert_valid_model_schema_generation(model)
        mime_documents = prepare_mime_document_list(documents)
        loaded_json_schema = load_json_schema(json_schema)
        data = {
            "json_schema": loaded_json_schema,
            "documents": [doc.model_dump() for doc in mime_documents],
            "ground_truths": ground_truths,
            "instructions": instructions if instructions else None,
            "model": model,
            "temperature": temperature,
            "modality": modality,
            "flat_likelihoods": flat_likelihoods,
            "tools_config": tools_config.model_dump(),
        }
        EnhanceSchemaRequest.model_validate(data)
        return PreparedRequest(method="POST", url="/v1/schemas/enhance", data=data)

    def prepare_get(self, schema_id: str) -> PreparedRequest:
        return PreparedRequest(method="GET", url=f"/v1/schemas/{schema_id}")


class Schemas(SyncAPIResource, SchemasMixin):
    def load(self, json_schema: dict[str, Any] | Path | str | None = None, pydantic_model: type[BaseModel] | None = None) -> Schema:
        """Load a schema from a JSON schema.

        Args:
            json_schema: The JSON schema to load
        """
        if json_schema:
            return Schema(json_schema=load_json_schema(json_schema))
        elif pydantic_model:
            return Schema(pydantic_model=pydantic_model)
        else:
            raise ValueError("Either json_schema or pydantic_model must be provided")

    def generate(
        self,
        documents: Sequence[Path | str | bytes | MIMEData | IOBase | PIL.Image.Image],
        instructions: str | None = None,
        model: str = "gpt-4o-2024-11-20",
        temperature: float = 0,
        modality: Modality = "native",
    ) -> Schema:
        """
        Generate a complete JSON schema by analyzing the provided documents.

        The generated schema includes X-Prompts for enhanced LLM interactions:
        - X-SystemPrompt: Defines high-level instructions and context for consistent LLM behavior
        - X-ReasoningPrompt: Creates auxiliary reasoning fields for complex data processing

        Args:
            documents: List of documents (as MIMEData) to analyze

        Returns:
            dict[str, Any]: Generated JSON schema with X-Prompts based on document analysis

        Raises:
            HTTPException if the request fails
        """

        prepared_request = self.prepare_generate(documents, instructions, model, temperature, modality)
        response = self._client._prepared_request(prepared_request)
        return Schema.model_validate(response)

    def evaluate(
        self,
        documents: Sequence[Path | str | bytes | MIMEData | IOBase | PIL.Image.Image],
        json_schema: dict[str, Any],
        ground_truths: list[dict[str, Any]] | None = None,
        model: str = "gpt-4o-mini",
        temperature: float = 0.0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        modality: Modality = "native",
        image_resolution_dpi: int = 96,
        browser_canvas: str = "A4",
        n_consensus: int = 1,
    ) -> EvaluateSchemaResponse:
        """
        Evaluate a schema by performing extractions on provided documents.
        If ground truths are provided, compares extractions against them.
        Otherwise, uses consensus likelihoods as metrics.

        Args:
            documents: List of documents to evaluate against
            json_schema: The JSON schema to evaluate
            ground_truths: Optional list of ground truth dictionaries to compare against
            model: The model to use for extraction
            temperature: The temperature to use for extraction
            reasoning_effort: The reasoning effort to use for extraction
            modality: The modality to use for extraction
            image_resolution_dpi: The DPI of the image. Defaults to 96.
            browser_canvas: The canvas size of the browser. Must be one of:
                - "A3" (11.7in x 16.54in)
                - "A4" (8.27in x 11.7in) 
                - "A5" (5.83in x 8.27in)
                Defaults to "A4".
            n_consensus: Number of consensus rounds to perform

        Returns:
            EvaluateSchemaResponse: Dictionary containing evaluation metrics for each document.
            Each metric includes:
            - similarity: Average similarity/likelihood score
            - similarities: Per-field similarity/likelihood scores
            - flat_similarities: Flattened per-field similarity/likelihood scores
            - aligned_* versions of the above metrics

        Raises:
            ValueError: If ground_truths is provided and its length doesn't match documents
            HTTPException: If the request fails or if there are too many documents
        """
        prepared_request = self.prepare_evaluate(
            documents=documents,
            json_schema=json_schema,
            ground_truths=ground_truths,
            model=model,
            temperature=temperature,
            reasoning_effort=reasoning_effort,
            modality=modality,
            image_resolution_dpi=image_resolution_dpi,
            browser_canvas=browser_canvas,
            n_consensus=n_consensus,
        )
        response = self._client._prepared_request(prepared_request)
        return EvaluateSchemaResponse.model_validate(response)

    def enhance(
        self,
        json_schema: dict[str, Any] | Path | str,
        documents: Sequence[Path | str | bytes | MIMEData | IOBase | PIL.Image.Image],
        ground_truths: list[dict[str, Any]] | None = None,
        instructions: str | None = None,
        model: str = "gpt-4o-2024-11-20",
        temperature: float = 0,
        modality: Modality = "native",
        flat_likelihoods: list[dict[str, float]] | dict[str, float] | None = None,
        tools_config: EnhanceSchemaConfigDict | None = None,
    ) -> Schema:
        prepared_request = self.prepare_enhance(
            json_schema, documents, ground_truths, instructions, model, temperature, modality, flat_likelihoods, EnhanceSchemaConfig.model_validate(tools_config or {})
        )
        response = self._client._prepared_request(prepared_request)
        return Schema(json_schema=response["json_schema"])


class AsyncSchemas(AsyncAPIResource, SchemasMixin):
    async def load(self, json_schema: dict[str, Any] | Path | str | None = None, pydantic_model: type[BaseModel] | None = None) -> Schema:
        """Load a schema from a JSON schema.

        Args:
            json_schema: The JSON schema to load
            pydantic_model: The Pydantic model to load
        """
        if json_schema:
            return Schema(json_schema=load_json_schema(json_schema))
        elif pydantic_model:
            return Schema(pydantic_model=pydantic_model)
        else:
            raise ValueError("Either json_schema or pydantic_model must be provided")

    """Schemas Asyncronous API wrapper"""

    async def get(self, schema_id: str) -> Schema:
        """Retrieve a schema by ID.

        Args:
            schema_id: The ID of the schema to retrieve

        Returns:
            Schema: The retrieved schema object
        """

        prepared_request = self.prepare_get(schema_id)
        response = await self._client._prepared_request(prepared_request)
        return Schema.model_validate(response)

    async def generate(
        self,
        documents: Sequence[Path | str | bytes | MIMEData | IOBase | PIL.Image.Image],
        instructions: str | None = None,
        model: str = "gpt-4o-2024-11-20",
        temperature: float = 0.0,
        modality: Modality = "native",
    ) -> Schema:
        """
        Generate a complete JSON schema by analyzing the provided documents.

        The generated schema includes X-Prompts for enhanced LLM interactions:
        - X-SystemPrompt: Defines high-level instructions and context for consistent LLM behavior
        - X-FieldPrompt: Enhances standard description fields with specific extraction guidance
        - X-ReasoningPrompt: Creates auxiliary reasoning fields for complex data processing

        Args:
            documents: List of documents (as MIMEData) to analyze

        Returns:
            dict[str, Any]: Generated JSON schema with X-Prompts based on document analysis

        Raises
            HTTPException if the request fails
        """
        prepared_request = self.prepare_generate(documents, instructions, model, temperature, modality)
        response = await self._client._prepared_request(prepared_request)
        return Schema.model_validate(response)

    async def evaluate(
        self,
        documents: Sequence[Path | str | bytes | MIMEData | IOBase | PIL.Image.Image],
        json_schema: dict[str, Any],
        ground_truths: list[dict[str, Any]] | None = None,
        model: str = "gpt-4o-mini",
        temperature: float = 0.0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        modality: Modality = "native",
        image_resolution_dpi: int = 96,
        browser_canvas: str = "A4",
        n_consensus: int = 1,
    ) -> EvaluateSchemaResponse:
        """
        Evaluate a schema by performing extractions on provided documents.
        If ground truths are provided, compares extractions against them.
        Otherwise, uses consensus likelihoods as metrics.

        Args:
            documents: List of documents to evaluate against
            json_schema: The JSON schema to evaluate
            ground_truths: Optional list of ground truth dictionaries to compare against
            model: The model to use for extraction
            temperature: The temperature to use for extraction
            reasoning_effort: The reasoning effort to use for extraction
            modality: The modality to use for extraction
            image_resolution_dpi: The DPI of the image. Defaults to 96.
            browser_canvas: The canvas size of the browser. Must be one of:
                - "A3" (11.7in x 16.54in)
                - "A4" (8.27in x 11.7in) 
                - "A5" (5.83in x 8.27in)
                Defaults to "A4".
            n_consensus: Number of consensus rounds to perform

        Returns:
            EvaluateSchemaResponse: Dictionary containing evaluation metrics for each document.
            Each metric includes:
            - similarity: Average similarity/likelihood score
            - similarities: Per-field similarity/likelihood scores
            - flat_similarities: Flattened per-field similarity/likelihood scores
            - aligned_* versions of the above metrics

        Raises:
            ValueError: If ground_truths is provided and its length doesn't match documents
            HTTPException: If the request fails or if there are too many documents
        """
        prepared_request = self.prepare_evaluate(
            documents=documents,
            json_schema=json_schema,
            ground_truths=ground_truths,
            model=model,
            temperature=temperature,
            reasoning_effort=reasoning_effort,
            modality=modality,
            image_resolution_dpi=image_resolution_dpi,
            browser_canvas=browser_canvas,
            n_consensus=n_consensus,
        )
        response = await self._client._prepared_request(prepared_request)
        return EvaluateSchemaResponse.model_validate(response)

    async def enhance(
        self,
        json_schema: dict[str, Any] | Path | str,
        documents: Sequence[Path | str | bytes | MIMEData | IOBase | PIL.Image.Image],
        ground_truths: list[dict[str, Any]] | None = None,
        instructions: str | None = None,
        model: str = "gpt-4o-2024-11-20",
        temperature: float = 0,
        modality: Modality = "native",
        flat_likelihoods: list[dict[str, float]] | dict[str, float] | None = None,
        tools_config: EnhanceSchemaConfigDict | None = None,
    ) -> Schema:
        prepared_request = self.prepare_enhance(
            json_schema, documents, ground_truths, instructions, model, temperature, modality, flat_likelihoods, EnhanceSchemaConfig.model_validate(tools_config or {})
        )
        response = await self._client._prepared_request(prepared_request)
        return Schema(json_schema=response["json_schema"])



import base64
import datetime
from typing import Any, Literal, Optional

from anthropic.types.message import Message
from anthropic.types.message_param import MessageParam
from openai.types.chat import ChatCompletionMessageParam
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from openai.types.chat.chat_completion_chunk import Choice as ChoiceChunk
from openai.types.chat.chat_completion_chunk import ChoiceDelta as ChoiceDeltaChunk
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from openai.types.chat.parsed_chat_completion import ParsedChatCompletion, ParsedChoice
from openai.types.responses.response import Response
from openai.types.responses.response_input_param import ResponseInputItemParam
from pydantic import BaseModel, ConfigDict, Field, ValidationInfo, computed_field, field_validator, model_validator

from ..._utils.usage.usage import compute_cost_from_model, compute_cost_from_model_with_breakdown, CostBreakdown

from ..._utils.ai_models import find_provider_from_model
from ..ai_models import AIProvider, Amount, get_model_card
from ..chat import ChatCompletionUiformMessage
from ..mime import MIMEData
from ..modalities import Modality
from ..standards import ErrorDetail, StreamingBaseModel


class DocumentExtractRequest(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    document: MIMEData = Field(..., description="Document to be analyzed")
    modality: Modality
    image_resolution_dpi: int = Field(default=96, description="Resolution of the image sent to the LLM")
    browser_canvas: Literal['A3', 'A4', 'A5'] = Field(
        default='A4', description="Sets the size of the browser canvas for rendering documents in browser-based processing. Choose a size that matches the document type."
    )
    model: str = Field(..., description="Model used for chat completion")
    json_schema: dict[str, Any] = Field(..., description="JSON schema format used to validate the output data.")
    temperature: float = Field(default=0.0, description="Temperature for sampling. If not provided, the default temperature for the model will be used.", examples=[0.0])
    reasoning_effort: ChatCompletionReasoningEffort = Field(
        default="medium", description="The effort level for the model to reason about the input data. If not provided, the default reasoning effort for the model will be used."
    )
    n_consensus: int = Field(default=1, description="Number of consensus models to use for extraction. If greater than 1 the temperature cannot be 0.")
    # Regular fields
    stream: bool = Field(default=False, description="If true, the extraction will be streamed to the user using the active WebSocket connection")
    seed: int | None = Field(default=None, description="Seed for the random number generator. If not provided, a random seed will be generated.", examples=[None])
    store: bool = Field(default=True, description="If true, the extraction will be stored in the database")
    need_validation: bool = Field(default=False, description="If true, the extraction will be validated against the schema")

    # Add a model validator that rejects n_consensus > 1 if temperature is 0
    @field_validator("n_consensus")
    def check_n_consensus(cls, v: int, info: ValidationInfo) -> int:
        if v > 1 and info.data.get("temperature") == 0:
            raise ValueError("n_consensus greater than 1 but temperature is 0")
        return v


class ConsensusModel(BaseModel):
    model: str = Field(description="Model name")
    temperature: float = Field(default=0.0, description="Temperature for consensus")
    reasoning_effort: ChatCompletionReasoningEffort = Field(
        default="medium", description="The effort level for the model to reason about the input data. If not provided, the default reasoning effort for the model will be used."
    )


# For location of fields in the document (OCR)
class FieldLocation(BaseModel):
    label: str = Field(..., description="The label of the field")
    value: str = Field(..., description="The extracted value of the field")
    quote: str = Field(..., description="The quote of the field (verbatim from the document)")
    file_id: str | None = Field(default=None, description="The ID of the file")
    page: int | None = Field(default=None, description="The page number of the field (1-indexed)")
    bboxes_normalized: list[tuple[float, float, float, float]] | None = Field(default=None, description="The normalized bounding boxes of the field")
    score: float | None = Field(default=None, description="The score of the field")
    match_level: Literal["token", "line", "block"] | None = Field(default=None, description="The level of the match (token, line, block)")


class UiParsedChoice(ParsedChoice):
    # Adaptable ParsedChoice that allows None for the finish_reason
    finish_reason: Literal["stop", "length", "tool_calls", "content_filter", "function_call"] | None = None  # type: ignore
    field_locations: dict[str, list[FieldLocation]] | None = Field(default=None, description="The locations of the fields in the document, if available")
    key_mapping: dict[str, Optional[str]] | None = Field(default=None, description="Mapping of consensus keys to original model keys")


LikelihoodsSource = Literal["consensus", "log_probs"]


class UiParsedChatCompletion(ParsedChatCompletion):
    extraction_id: str | None = None
    choices: list[UiParsedChoice]
    # Additional metadata fields (UIForm)
    likelihoods: Optional[dict[str, Any]] = Field(
        default=None, description="Object defining the uncertainties of the fields extracted when using consensus. Follows the same structure as the extraction object."
    )
    schema_validation_error: ErrorDetail | None = None
    # Timestamps
    request_at: datetime.datetime | None = Field(default=None, description="Timestamp of the request")
    first_token_at: datetime.datetime | None = Field(default=None, description="Timestamp of the first token of the document. If non-streaming, set to last_token_at")
    last_token_at: datetime.datetime | None = Field(default=None, description="Timestamp of the last token of the document")

    @computed_field
    @property
    def api_cost(self) -> Optional[Amount]:
        if self.usage:
            try:
                cost = compute_cost_from_model(self.model, self.usage)
                return cost
            except Exception as e:
                print(f"Error computing cost: {e}")
                return None
        return None


class UiResponse(Response):
    extraction_id: str | None = None
    # Additional metadata fields (UIForm)
    likelihoods: Optional[dict[str, Any]] = Field(
        default=None, description="Object defining the uncertainties of the fields extracted when using consensus. Follows the same structure as the extraction object."
    )
    schema_validation_error: ErrorDetail | None = None
    # Timestamps
    request_at: datetime.datetime | None = Field(default=None, description="Timestamp of the request")
    first_token_at: datetime.datetime | None = Field(default=None, description="Timestamp of the first token of the document. If non-streaming, set to last_token_at")
    last_token_at: datetime.datetime | None = Field(default=None, description="Timestamp of the last token of the document")


class LogExtractionRequest(BaseModel):
    messages: list[ChatCompletionUiformMessage] | None = None  # TODO: compatibility with Anthropic
    openai_messages: list[ChatCompletionMessageParam] | None = None
    openai_responses_input: list[ResponseInputItemParam] | None = None
    anthropic_messages: list[MessageParam] | None = None
    anthropic_system_prompt: str | None = None
    document: MIMEData = Field(
        default=MIMEData(
            filename="dummy.txt",
            # url is a base64 encoded string with the mime type and the content. For the dummy one we will send a .txt file with the text "No document provided"
            url="data:text/plain;base64," + base64.b64encode(b"No document provided").decode("utf-8"),
        ),
        description="Document analyzed, if not provided a dummy one will be created with the text 'No document provided'",
    )
    completion: dict | UiParsedChatCompletion | Message | ParsedChatCompletion | ChatCompletion | None = None
    openai_responses_output: Response | None = None
    json_schema: dict[str, Any]
    model: str
    temperature: float

    # Validate that at least one of the messages, openai_messages, anthropic_messages is provided using model_validator
    @model_validator(mode="before")
    def validation(cls, data: Any) -> Any:
        messages_candidates = [data.get("messages"), data.get("openai_messages"), data.get("anthropic_messages"), data.get("openai_responses_input")]
        messages_candidates = [candidate for candidate in messages_candidates if candidate is not None]
        if len(messages_candidates) != 1:
            raise ValueError("Exactly one of the messages, openai_messages, anthropic_messages, openai_responses_input must be provided")

        # Validate that if anthropic_messages is provided, anthropic_system_prompt is also provided
        if data.get("anthropic_messages") is not None and data.get("anthropic_system_prompt") is None:
            raise ValueError("anthropic_system_prompt must be provided if anthropic_messages is provided")

        completion_candidates = [data.get("completion"), data.get("openai_responses_output")]
        completion_candidates = [candidate for candidate in completion_candidates if candidate is not None]
        if len(completion_candidates) != 1:
            raise ValueError("Exactly one of completion, openai_responses_output must be provided")

        return data


class LogExtractionResponse(BaseModel):
    extraction_id: str | None = None  # None only in case of error
    status: Literal["success", "error"]
    error_message: str | None = None


# DocumentExtractResponse = UiParsedChatCompletion


###### I'll place here for now -- New Streaming API


# We build from the openai.types.chat.chat_completion_chunk.ChatCompletionChunk adding just two three additional fields:
# - is_valid_json: list[bool]               #  Whether the total accumulated content is a valid JSON
# - likelihoods: dict[str, float]     #  The delta of the flattened likelihoods (to be merged with the cumulated likelihoods)
# - schema_validation_error: ErrorDetail | None = None #  The error in the schema validation of the total accumulated content


class UiParsedChoiceDeltaChunk(ChoiceDeltaChunk):
    flat_likelihoods: dict[str, float] = {}
    flat_parsed: dict[str, Any] = {}
    flat_deleted_keys: list[str] = []
    field_locations: dict[str, list[FieldLocation]] | None = Field(default=None, description="The locations of the fields in the document, if available")
    is_valid_json: bool = False
    key_mapping: dict[str, Optional[str]] | None = Field(default=None, description="Mapping of consensus keys to original model keys")


class UiParsedChoiceChunk(ChoiceChunk):
    delta: UiParsedChoiceDeltaChunk


class UiParsedChatCompletionChunk(StreamingBaseModel, ChatCompletionChunk):
    extraction_id: str | None = None
    choices: list[UiParsedChoiceChunk]
    schema_validation_error: ErrorDetail | None = None
    # Timestamps
    request_at: datetime.datetime | None = Field(default=None, description="Timestamp of the request")
    first_token_at: datetime.datetime | None = Field(default=None, description="Timestamp of the first token of the document. If non-streaming, set to last_token_at")
    last_token_at: datetime.datetime | None = Field(default=None, description="Timestamp of the last token of the document")

    @computed_field
    @property
    def api_cost(self) -> Optional[Amount]:
        if self.usage:
            try:
                cost = compute_cost_from_model(self.model, self.usage)
                return cost
            except Exception as e:
                print(f"Error computing cost: {e}")
                return None
        return None

    @computed_field  # type: ignore
    @property
    def cost_breakdown(self) -> Optional[CostBreakdown]:
        if self.usage:
            try:
                cost = compute_cost_from_model_with_breakdown(self.model, self.usage)
                return cost
            except Exception as e:
                print(f"Error computing cost: {e}")
                return None
        return None

    def chunk_accumulator(self, previous_cumulated_chunk: "UiParsedChatCompletionChunk | None" = None) -> "UiParsedChatCompletionChunk":
        """
        Accumulate the chunk into the state, returning a new UiParsedChatCompletionChunk with the accumulated content that could be yielded alone to generate the same state.
        """

        def safe_get_delta(chnk: "UiParsedChatCompletionChunk | None", index: int) -> UiParsedChoiceDeltaChunk:
            if chnk is not None and index < len(chnk.choices):
                return chnk.choices[index].delta
            else:
                return UiParsedChoiceDeltaChunk(
                    content="",
                    flat_parsed={},
                    flat_likelihoods={},
                    is_valid_json=False,
                )

        max_choices = max(len(self.choices), len(previous_cumulated_chunk.choices)) if previous_cumulated_chunk is not None else len(self.choices)

        # Get the current chunk missing content, flat_deleted_keys and is_valid_json
        acc_flat_deleted_keys = [safe_get_delta(self, i).flat_deleted_keys for i in range(max_choices)]
        acc_is_valid_json = [safe_get_delta(self, i).is_valid_json for i in range(max_choices)]
        acc_field_locations = [safe_get_delta(self, i).field_locations for i in range(max_choices)]  # This is only present in the last chunk.
        # Delete from previous_cumulated_chunk.choices[i].delta.flat_parsed the keys that are in safe_get_delta(self, i).flat_deleted_keys
        for i in range(max_choices):
            previous_delta = safe_get_delta(previous_cumulated_chunk, i)
            current_delta = safe_get_delta(self, i)
            for deleted_key in current_delta.flat_deleted_keys:
                previous_delta.flat_parsed.pop(deleted_key, None)
                previous_delta.flat_likelihoods.pop(deleted_key, None)
        # Accumulate the flat_parsed and flat_likelihoods
        acc_flat_parsed = [safe_get_delta(previous_cumulated_chunk, i).flat_parsed | safe_get_delta(self, i).flat_parsed for i in range(max_choices)]
        acc_flat_likelihoods = [safe_get_delta(previous_cumulated_chunk, i).flat_likelihoods | safe_get_delta(self, i).flat_likelihoods for i in range(max_choices)]
        acc_key_mapping = [safe_get_delta(previous_cumulated_chunk, i).key_mapping or safe_get_delta(self, i).key_mapping for i in range(max_choices)]

        acc_content = [(safe_get_delta(previous_cumulated_chunk, i).content or "") + (safe_get_delta(self, i).content or "") for i in range(max_choices)]
        usage = self.usage
        first_token_at = self.first_token_at
        last_token_at = self.last_token_at
        request_at = self.request_at

        return UiParsedChatCompletionChunk(
            extraction_id=self.extraction_id,
            id=self.id,
            created=self.created,
            model=self.model,
            object=self.object,
            usage=usage,
            choices=[
                UiParsedChoiceChunk(
                    delta=UiParsedChoiceDeltaChunk(
                        content=acc_content[i],
                        flat_parsed=acc_flat_parsed[i],
                        flat_likelihoods=acc_flat_likelihoods[i],
                        flat_deleted_keys=acc_flat_deleted_keys[i],
                        field_locations=acc_field_locations[i],
                        is_valid_json=acc_is_valid_json[i],
                        key_mapping=acc_key_mapping[i],
                    ),
                    index=i,
                )
                for i in range(max_choices)
            ],
            schema_validation_error=self.schema_validation_error,
            request_at=request_at,
            first_token_at=first_token_at,
            last_token_at=last_token_at,
        )



from typing import Any

from ..._resource import AsyncAPIResource, SyncAPIResource
from .external_api_keys import AsyncExternalAPIKeys, ExternalAPIKeys


class Secrets(SyncAPIResource):
    """Automations API wrapper"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.external_api_keys = ExternalAPIKeys(client=client)


class AsyncSecrets(AsyncAPIResource):
    """Automations API wrapper"""

    def __init__(self, client: Any) -> None:
        super().__init__(client=client)
        self.external_api_keys = AsyncExternalAPIKeys(client=client)



from typing import Any, Dict, Literal, Optional

from ...._resource import SyncAPIResource, AsyncAPIResource
from ....types.logs import AutomationLog, ExternalRequestLog, ListLogs
from ....types.standards import PreparedRequest


class LogsMixin:
    def prepare_get(self, log_id: str) -> PreparedRequest:
        """Get a specific automation log by ID.

        Args:
            log_id: ID of the log to retrieve

        Returns:
            PreparedRequest: The prepared request
        """
        return PreparedRequest(method="GET", url=f"/v1/processors/automations/logs/{log_id}")

    def prepare_list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        # Filtering parameters
        status_code: Optional[int] = None,
        status_class: Optional[Literal["2xx", "3xx", "4xx", "5xx"]] = None,
        automation_id: Optional[str] = None,
        webhook_url: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> PreparedRequest:
        """List automation logs with pagination support.

        Args:
            before: Optional cursor for pagination before a specific log ID
            after: Optional cursor for pagination after a specific log ID
            limit: Optional limit on number of results (max 100)
            order: Optional sort order ("asc" or "desc")
            status_code: Optional filter by status code
            status_class: Optional filter by status_class
            automation_id: Optional filter by deployment ID
            webhook_url: Optional filter by webhook URL
            schema_id: Optional filter by schema ID
            schema_data_id: Optional filter by schema data ID

        Returns:
            PreparedRequest: The prepared request
        """
        params = {
            "before": before,
            "after": after,
            "limit": limit,
            "order": order,
            "status_code": status_code,
            "status_class": status_class,
            "automation_id": automation_id,
            "webhook_url": webhook_url,
            "schema_id": schema_id,
            "schema_data_id": schema_data_id,
        }
        # Remove None values
        params = {k: v for k, v in params.items() if v is not None}

        return PreparedRequest(method="GET", url="/v1/processors/automations/logs", params=params)

    def prepare_rerun(self, log_id: str) -> PreparedRequest:
        """Rerun a webhook from an existing AutomationLog.

        Args:
            log_id: ID of the log to rerun

        Returns:
            PreparedRequest: The prepared request
        """
        return PreparedRequest(method="POST", url=f"/v1/processors/automations/logs/{log_id}/rerun")


class Logs(SyncAPIResource, LogsMixin):
    """Logs API wrapper for managing automation logs"""

    def get(self, log_id: str) -> AutomationLog:
        """Get a specific automation log by ID.

        Args:
            log_id: ID of the log to retrieve

        Returns:
            AutomationLog: The automation log
        """
        request = self.prepare_get(log_id)
        response = self._client._prepared_request(request)
        return AutomationLog.model_validate(response)

    def list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        status_code: Optional[int] = None,
        status_class: Optional[Literal["2xx", "3xx", "4xx", "5xx"]] = None,
        automation_id: Optional[str] = None,
        webhook_url: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> ListLogs:
        """List automation logs with pagination support.

        Args:
            before: Optional cursor for pagination before a specific log ID
            after: Optional cursor for pagination after a specific log ID
            limit: Optional limit on number of results (max 100)
            order: Optional sort order ("asc" or "desc")
            status_code: Optional filter by status code
            status_class: Optional filter by status_class
            automation_id: Optional filter by deployment ID
            webhook_url: Optional filter by webhook URL
            schema_id: Optional filter by schema ID
            schema_data_id: Optional filter by schema data ID

        Returns:
            ListLogs: Paginated list of automation logs with metadata
        """
        request = self.prepare_list(before, after, limit, order, status_code, status_class, automation_id, webhook_url, schema_id, schema_data_id)
        response = self._client._prepared_request(request)
        return ListLogs.model_validate(response)

    def rerun(self, log_id: str) -> ExternalRequestLog:
        """Rerun a webhook from an existing AutomationLog.

        Args:
            log_id: ID of the log to rerun

        Returns:
            ExternalRequestLog: The result of the rerun webhook call
        """
        request = self.prepare_rerun(log_id)
        response = self._client._prepared_request(request)

        print(f"Webhook call run successfully. Log available at https://docs.uiform.com/dashboard/processors/logs/{log_id}")

        return ExternalRequestLog.model_validate(response)


class AsyncLogs(AsyncAPIResource, LogsMixin):
    """Async Logs API wrapper for managing automation logs"""

    async def get(self, log_id: str) -> AutomationLog:
        """Get a specific automation log by ID.

        Args:
            log_id: ID of the log to retrieve

        Returns:
            AutomationLog: The automation log
        """
        request = self.prepare_get(log_id)
        response = await self._client._prepared_request(request)
        return AutomationLog.model_validate(response)

    async def list(
        self,
        before: Optional[str] = None,
        after: Optional[str] = None,
        limit: Optional[int] = 10,
        order: Optional[Literal["asc", "desc"]] = "desc",
        status_code: Optional[int] = None,
        status_class: Optional[Literal["2xx", "3xx", "4xx", "5xx"]] = None,
        automation_id: Optional[str] = None,
        webhook_url: Optional[str] = None,
        schema_id: Optional[str] = None,
        schema_data_id: Optional[str] = None,
    ) -> ListLogs:
        """List automation logs with pagination support.

        Args:
            before: Optional cursor for pagination before a specific log ID
            after: Optional cursor for pagination after a specific log ID
            limit: Optional limit on number of results (max 100)
            order: Optional sort order ("asc" or "desc")
            status_code: Optional filter by status code
            status_class: Optional filter by status_class
            automation_id: Optional filter by deployment ID
            webhook_url: Optional filter by webhook URL
            schema_id: Optional filter by schema ID
            schema_data_id: Optional filter by schema data ID

        Returns:
            ListLogs: Paginated list of automation logs with metadata
        """
        request = self.prepare_list(before, after, limit, order, status_code, status_class, automation_id, webhook_url, schema_id, schema_data_id)
        response = await self._client._prepared_request(request)
        return ListLogs.model_validate(response)

    async def rerun(self, log_id: str) -> ExternalRequestLog:
        """Rerun a webhook from an existing AutomationLog.

        Args:
            log_id: ID of the log to rerun

        Returns:
            ExternalRequestLog: The result of the rerun webhook call
        """
        request = self.prepare_rerun(log_id)
        response = await self._client._prepared_request(request)

        print(f"Webhook call run successfully. Log available at https://docs.uiform.com/dashboard/processors/logs/{log_id}")

        return ExternalRequestLog.model_validate(response)



import json
from pathlib import Path
from typing import Any, AsyncGenerator, Generator

from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from openai.types.chat.parsed_chat_completion import ParsedChatCompletionMessage
from openai.types.shared_params.response_format_json_schema import ResponseFormatJSONSchema
#from openai.lib._parsing import ResponseFormatT
from pydantic import BaseModel as ResponseFormatT


from ..._resource import AsyncAPIResource, SyncAPIResource
from ..._utils.ai_models import assert_valid_model_extraction
from ..._utils.json_schema import load_json_schema, unflatten_dict
from ..._utils.stream_context_managers import as_async_context_manager, as_context_manager
from ...types.chat import ChatCompletionUiformMessage
from ...types.completions import UiChatCompletionsRequest
from ...types.documents.extractions import UiParsedChatCompletion, UiParsedChatCompletionChunk, UiParsedChoice
from ...types.standards import PreparedRequest
from ...types.schemas.object import Schema



class BaseCompletionsMixin:

    def prepare_parse(
        self,
        response_format: type[ResponseFormatT],
        messages: list[ChatCompletionUiformMessage],
        model: str,
        temperature: float,
        reasoning_effort: ChatCompletionReasoningEffort,
        stream: bool,
        n_consensus: int,
        idempotency_key: str | None = None,
    ) -> PreparedRequest:
        assert_valid_model_extraction(model)

        json_schema = response_format.model_json_schema()

        schema_obj = Schema(json_schema=json_schema)

        data = {
            "messages": messages,
            "response_format": {
                "type": "json_schema",
                "json_schema": {
                    "name": schema_obj.id,
                    "schema": schema_obj.inference_json_schema,
                    "strict": True,
                },
            },
            "model": model,
            "temperature": temperature,
            "stream": stream,
            "reasoning_effort": reasoning_effort,
            "n_consensus": n_consensus,
        }

        # Validate DocumentAPIRequest data (raises exception if invalid)
        ui_chat_completions_request = UiChatCompletionsRequest.model_validate(data)

        return PreparedRequest(method="POST", url="/v1/completions", data=ui_chat_completions_request.model_dump(), idempotency_key=idempotency_key)


    def prepare_create(
        self,
        response_format: ResponseFormatJSONSchema,
        messages: list[ChatCompletionUiformMessage],
        model: str,
        temperature: float,
        reasoning_effort: ChatCompletionReasoningEffort,
        stream: bool,
        n_consensus: int,
        idempotency_key: str | None = None,
    ) -> PreparedRequest:
        
        json_schema = response_format["json_schema"].get("schema")

        assert isinstance(json_schema, dict), f"json_schema must be a dictionary, got {type(json_schema)}"

        schema_obj = Schema(json_schema=json_schema)

        data = {
            "messages": messages,
            "response_format": {
                "type": "json_schema",
                "json_schema": {
                    "name": schema_obj.id,
                    "schema": schema_obj.inference_json_schema,
                    "strict": True,
                },
            },
            "model": model,
            "temperature": temperature,
            "stream": stream,
            "reasoning_effort": reasoning_effort,
            "n_consensus": n_consensus,
        }

        # Validate DocumentAPIRequest data (raises exception if invalid)
        ui_chat_completions_request = UiChatCompletionsRequest.model_validate(data)

        return PreparedRequest(method="POST", url="/v1/completions", data=ui_chat_completions_request.model_dump(), idempotency_key=idempotency_key)


class Completions(SyncAPIResource, BaseCompletionsMixin):
    """Multi-provider Completions API wrapper"""

    def create(
        self,
        response_format: ResponseFormatJSONSchema,
        messages: list[ChatCompletionUiformMessage],
        model: str = "gpt-4o-2024-08-06",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        n_consensus: int = 1,
        idempotency_key: str | None = None,
        stream: bool = False,
    ) -> UiParsedChatCompletion:
        """
        Create a completion using the UiForm API.
        """

        request = self.prepare_create(
            model=model,
            temperature=temperature,
            reasoning_effort=reasoning_effort,
            stream=stream,
            messages=messages,
            response_format=response_format,
            n_consensus=n_consensus,
            idempotency_key=idempotency_key,
        )

        response = self._client._prepared_request(request)

        return UiParsedChatCompletion.model_validate(response)


    def parse(
        self,
        response_format: type[ResponseFormatT],
        messages: list[ChatCompletionUiformMessage],
        model: str = "gpt-4o-2024-08-06",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        n_consensus: int = 1,
        idempotency_key: str | None = None,
    ) -> UiParsedChatCompletion:
        """
        Parse messages using the UiForm API to extract structured data according to the provided JSON schema.

        Args:
            response_format: JSON schema defining the expected data structure
            messages: List of chat messages to parse
            model: The AI model to use for processing
            temperature: Model temperature setting (0-1)
            reasoning_effort: The effort level for the model to reason about the input data
            idempotency_key: Idempotency key for request
            store: Whether to store the data in the UiForm database

        Returns:
            UiParsedChatCompletion: Parsed response from the API
        """
        request = self.prepare_parse(
            response_format=response_format,
            messages=messages,
            model=model,
            temperature=temperature,
            reasoning_effort=reasoning_effort,
            stream=False,
            n_consensus=n_consensus,
            idempotency_key=idempotency_key,
        )
        response = self._client._prepared_request(request)

        return UiParsedChatCompletion.model_validate(response)


class AsyncCompletions(AsyncAPIResource, BaseCompletionsMixin):
    """Multi-provider Completions API wrapper for asynchronous usage."""

    async def create(
        self,
        response_format: ResponseFormatJSONSchema,
        messages: list[ChatCompletionUiformMessage],
        model: str = "gpt-4o-2024-08-06",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        n_consensus: int = 1,
        idempotency_key: str | None = None,
        stream: bool = False,
    ) -> UiParsedChatCompletion:
        """
        Create a completion using the UiForm API.
        """

        request = self.prepare_create(
            model=model,
            temperature=temperature,
            reasoning_effort=reasoning_effort,
            stream=stream,
            messages=messages,
            response_format=response_format,
            n_consensus=n_consensus,
            idempotency_key=idempotency_key,
        )

        response = await self._client._prepared_request(request)
        return UiParsedChatCompletion.model_validate(response)




    async def parse(
        self,
        response_format: type[ResponseFormatT],
        messages: list[ChatCompletionUiformMessage],
        model: str = "gpt-4o-2024-08-06",
        temperature: float = 0,
        reasoning_effort: ChatCompletionReasoningEffort = "medium",
        n_consensus: int = 1,
        idempotency_key: str | None = None,
    ) -> UiParsedChatCompletion:
        """
        Parse messages using the UiForm API asynchronously.

        Args:
            json_schema: JSON schema defining the expected data structure
            messages: List of chat messages to parse
            model: The AI model to use
            temperature: Model temperature setting (0-1)
            reasoning_effort: The effort level for the model to reason about the input data
            n_consensus: Number of consensus models to use for extraction
            idempotency_key: Idempotency key for request

        Returns:
            UiParsedChatCompletion: Parsed response from the API
        """
        request = self.prepare_parse(
            response_format=response_format,
            messages=messages,
            model=model,
            temperature=temperature,
            reasoning_effort=reasoning_effort,
            stream=False,
            n_consensus=n_consensus,
            idempotency_key=idempotency_key,
        )
        response = await self._client._prepared_request(request)
        return UiParsedChatCompletion.model_validate(response)



import datetime
import re
import shutil

# The goal is to leverage this piece of code to open a jsonl file and get an analysis of the performance of the model using a one-liner.
############# BENCHMARKING MODELS #############
from itertools import zip_longest
from typing import Any, Callable, Literal, Optional

import pandas as pd  # type: ignore
from Levenshtein import distance as levenshtein_distance
from pydantic import BaseModel

from ..types.db.annotations import AnnotationParameters


def normalize_string(text: str) -> str:
    """
    Normalize a string by removing non-alphanumeric characters and lowercasing.

    Args:
        text: Input string to normalize

    Returns:
        Normalized string with only alphanumeric characters, all lowercase
    """
    if not text:
        return ""
    # Remove all non-alphanumeric characters and convert to lowercase
    return re.sub(r'[^a-zA-Z0-9]', '', text).lower()


def hamming_distance_padded(s: str, t: str) -> int:
    """
    Compute the Hamming distance between two strings, treating spaces as wildcards.

    Args:
        s: The first string
        t: The second string

    Returns:
        The Hamming distance between the two strings
    """
    # Normalize inputs
    s = normalize_string(s)
    t = normalize_string(t)

    return sum(a != b for a, b in zip_longest(s, t, fillvalue=' '))


def hamming_similarity(str_1: str, str_2: str) -> float:
    """
    Compute the Hamming similarity between two strings.

    Args:
        str_1: The first string
        str_2: The second string

    Returns:
        A float between 0 and 1, where 1 means the strings are identical
    """
    # Normalize inputs
    str_1 = normalize_string(str_1)
    str_2 = normalize_string(str_2)

    max_length = max(len(str_1), len(str_2))

    if max_length == 0:
        return 1.0

    dist = hamming_distance_padded(str_1, str_2)
    return 1 - (dist / max_length)


def jaccard_similarity(str_1: str, str_2: str) -> float:
    """
    Compute the Jaccard similarity between two strings.

    Args:
        str_1: The first string
        str_2: The second string

    Returns:
        A float between 0 and 1, where 1 means the strings are identical
    """
    # Normalize inputs
    str_1 = normalize_string(str_1)
    str_2 = normalize_string(str_2)

    set_a = set(str_1)
    set_b = set(str_2)
    intersection = set_a & set_b
    union = set_a | set_b
    if not union:
        return 1.0
    return len(intersection) / len(union)


def levenshtein_similarity(str_1: str, str_2: str) -> float:
    """
    Calculate similarity between two values using Levenshtein distance.
    Returns a similarity score between 0.0 and 1.0.
    """
    # Normalize inputs
    str_1 = normalize_string(str_1)
    str_2 = normalize_string(str_2)

    max_length = max(len(str_1), len(str_2))

    if max_length == 0:
        return 1.0

    dist = levenshtein_distance(str_1, str_2)
    return 1 - (dist / max_length)


def key_normalization(key: str) -> str:
    """This method is useful to compare keys under list indexes (that refers to the same kind of error but on different list index position)"""
    # We will replace all .{i} with .* where i is the index of the list (using regex for this)
    key_parts = key.split(".")
    new_key_parts = []
    for key_part in key_parts:
        if key_part.isdigit():
            new_key_parts.append("*")
        else:
            new_key_parts.append(key_part)
    return ".".join(new_key_parts)


def compare_primitive_values(val1: str | int | float | bool | None, val2: str | int | float | bool | None, str_metric_function: Callable[[str, str], float]) -> float:
    # Handle leaf nodes (primitives) with type-specific comparisons
    # Special handling for None values
    # Both None means perfect match
    if val1 is None and val2 is None:
        return 1.0
    # One is None but the other has a value (False, "", 0, etc.)
    elif (val1 is None and not val2) or (val2 is None and not val1):
        return 1.0
    # One is None but the other has non None-compatible values (True, "string", 1, 1.5, etc.)
    elif val1 is None or val2 is None:
        return 0.0

    # From now on, we can assume that val1 and val2 are not None.
    # Type compatibility check
    if isinstance(val1, bool) and isinstance(val2, bool):
        return 1.0 if val1 is val2 else 0.0

    # Numeric comparison (int, float)
    if isinstance(val1, (int, float)) and isinstance(val2, (int, float)):
        # For numbers close to zero, use absolute difference
        if abs(val1) < 1e-4 and abs(val2) < 1e-4:
            return 1.0 if abs(val1 - val2) < 1e-4 else 0.0
        # Otherwise use relative difference
        max_val = max(abs(val1), abs(val2))
        return 1.0 - min(1.0, abs(val1 - val2) / max_val)

    # String comparison - use the provided metric function
    if isinstance(val1, str) and isinstance(val2, str):
        return float(str_metric_function(val1, val2))

    # If we get here, types are incompatible
    return 0.0


dictionary_metrics = Literal["levenshtein_similarity", "jaccard_similarity", "hamming_similarity"]


def compute_dict_difference(dict1: dict[str, Any], dict2: dict[str, Any], metric: dictionary_metrics) -> dict[str, Any]:
    """
    Compute the difference between two dictionaries recursively.

    Args:
        dict1: The first dictionary (can be nested)
        dict2: The second dictionary (can be nested)
        metric: The metric to use for comparison ("levenshtein_similarity", "jaccard_similarity", "hamming_similarity")

    Returns:
        A dictionary containing the difference between the two dictionaries
    """
    result: dict[str, Any] = {}

    if metric == "levenshtein_similarity":
        metric_function = levenshtein_similarity
    elif metric == "jaccard_similarity":
        metric_function = jaccard_similarity
    elif metric == "hamming_similarity":
        metric_function = hamming_similarity
    else:
        raise ValueError(f"Invalid metric: {metric}")

    def compare_values(val1: dict | list | tuple | str | int | float | bool | None, val2: dict | list | tuple | str | int | float | bool | None, path: str = "") -> Any:
        # If both are dictionaries, process recursively
        if isinstance(val1, dict) and isinstance(val2, dict):
            nested_result: dict[str, Any] = {}
            all_keys = set(val1.keys()) | set(val2.keys())

            for key in all_keys:
                norm_key = key_normalization(key)
                sub_val1 = val1.get(key, None)
                sub_val2 = val2.get(key, None)

                if sub_val1 is None or sub_val2 is None:
                    nested_result[norm_key] = None
                else:
                    nested_result[norm_key] = compare_values(sub_val1, sub_val2, f"{path}.{norm_key}" if path else norm_key)

            return nested_result

        # If both are lists/arrays, compare items with detailed results
        if isinstance(val1, (list, tuple)) and isinstance(val2, (list, tuple)):
            # If both lists are empty, they're identical
            if not val1 and not val2:
                return 1.0

            # Create a detailed element-by-element comparison
            array_result = {}
            similarities = []

            # Process each position in both arrays
            for i, (item1, item2) in enumerate(zip_longest(val1, val2, fillvalue=None)):
                element_key = str(i)  # Use index as dictionary key
                element_path = f"{path}.{i}" if path else str(i)

                if item1 is None or item2 is None:
                    # Handle lists of different lengths
                    array_result[element_key] = None
                    similarities.append(0.0)  # Penalize missing elements
                else:
                    # Compare the elements
                    comparison_result = compare_values(item1, item2, element_path)
                    array_result[element_key] = comparison_result

                    # Extract similarity metric for this element
                    if isinstance(comparison_result, dict):
                        # Calculate average from nested structure
                        numeric_values = [v for v in _extract_numeric_values(comparison_result) if v is not None]
                        if numeric_values:
                            similarities.append(sum(numeric_values) / len(numeric_values))
                    elif isinstance(comparison_result, (int, float)) and comparison_result is not None:
                        similarities.append(float(comparison_result))

            # Add overall similarity as a special key
            array_result["_similarity"] = sum(similarities) / max(len(similarities), 1) if similarities else 1.0

            return array_result

        # If one is a dict and the other isn't, return None
        if isinstance(val1, dict) or isinstance(val2, dict) or isinstance(val1, (list, tuple)) or isinstance(val2, (list, tuple)):
            return None

        # Handle leaf nodes (primitives) with type-specific comparisons
        return compare_primitive_values(val1, val2, metric_function)

    def _extract_numeric_values(d: dict) -> list[float]:
        """Extract all numeric values from a nested dictionary."""
        result = []
        for k, v in d.items():
            if isinstance(v, dict):
                # Recursively extract from nested dictionaries
                result.extend(_extract_numeric_values(v))
            elif isinstance(v, (int, float)) and not isinstance(v, bool):
                # Add numeric values
                result.append(v)
            # Skip non-numeric values
        return result

    # Normalize top-level keys
    dict1_normalized = {key_normalization(k): v for k, v in dict1.items()}
    dict2_normalized = {key_normalization(k): v for k, v in dict2.items()}

    # Process all keys from both dictionaries
    keys_intersect = set(dict1_normalized.keys()) & set(dict2_normalized.keys())
    keys_symmetric_difference = set(dict1_normalized.keys()) ^ set(dict2_normalized.keys())

    for key in keys_symmetric_difference:
        # When the key is not present in both dictionaries, we return None.
        result[key] = None

    for key in keys_intersect:
        # compare_values can handle None values, so we don't need to check for that.
        result[key] = compare_values(dict1_normalized[key], dict2_normalized[key], key)

    return result


def aggregate_dict_differences(dict_differences: list[dict[str, Any]]) -> tuple[dict[str, Any], dict[str, Any]]:
    """
    Aggregate a list of dictionary differences into a single dictionary with average values,
    handling nested dictionaries recursively.

    Args:
        dict_differences: A list of dictionaries containing similarity metrics (can be nested)

    Returns:
        A tuple containing:
        - A dictionary with the average similarity metrics across all input dictionaries
        - A dictionary with the uncertainty (standard deviation) for each metric
    """
    if not dict_differences:
        return {}, {}

    def aggregate_recursively(dicts_list: list[dict[str, Any]]) -> tuple[dict[str, Any], dict[str, Any]]:
        # Initialize result dictionaries
        result: dict[str, Any] = {}
        uncertainty: dict[str, Any] = {}

        # Collect all keys across all dictionaries
        all_keys: set[str] = set()
        for d in dicts_list:
            all_keys.update(d.keys())

        for key in all_keys:
            # Collect values for this key from all dictionaries
            values = []
            for d in dicts_list:
                if key in d:
                    values.append(d[key])

            # Skip if no valid values
            if not values:
                result[key] = None
                uncertainty[key] = None
                continue

            # Check if values are nested dictionaries
            if all(isinstance(v, dict) for v in values if v is not None):
                # Filter out None values
                nested_dicts = [v for v in values if v is not None]
                if nested_dicts:
                    nested_result, nested_uncertainty = aggregate_recursively(nested_dicts)
                    result[key] = nested_result
                    uncertainty[key] = nested_uncertainty
                else:
                    result[key] = None
                    uncertainty[key] = None
            else:
                # Handle leaf nodes (numeric values)
                numeric_values = [v for v in values if v is not None and isinstance(v, (int, float))]

                if numeric_values:
                    mean = sum(numeric_values) / len(numeric_values)
                    result[key] = mean

                    if len(numeric_values) > 1:
                        variance = sum((x - mean) ** 2 for x in numeric_values) / (len(numeric_values) - 1)
                        uncertainty[key] = max(0, variance) ** 0.5
                    else:
                        uncertainty[key] = 0.0
                else:
                    result[key] = None
                    uncertainty[key] = None

        return result, uncertainty

    return aggregate_recursively(dict_differences)


class SingleFileEval(BaseModel):
    """
    A class for evaluating metrics between two dictionaries.
    """

    evaluation_id: str
    file_id: str
    schema_id: str
    schema_data_id: str | None = None
    dict_1: dict[str, Any]
    dict_2: dict[str, Any]
    inference_settings_1: AnnotationParameters
    inference_settings_2: AnnotationParameters
    created_at: datetime.datetime
    organization_id: str
    hamming_similarity: dict[str, Any]
    jaccard_similarity: dict[str, Any]
    levenshtein_similarity: dict[str, Any]


class EvalMetric(BaseModel):
    average: dict[str, Any]
    std: dict[str, Any]


class EvalMetrics(BaseModel):
    schema_id: str
    distances: dict[dictionary_metrics, EvalMetric]


def flatten_dict(d: dict[str, Any], parent_key: str = '', sep: str = '.') -> dict[str, Any]:
    """Flatten a nested dictionary with dot-separated keys."""
    items: list[tuple[str, Any]] = []
    for k, v in d.items():
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def plot_metrics_with_uncertainty(analysis: dict[str, Any], uncertainties: Optional[dict[str, Any]] = None, top_n: int = 20, ascending: bool = False) -> None:
    """Plot a metric from analysis results using a horizontal bar chart with uncertainty.

    Args:
        analysis: Dictionary containing similarity scores (can be nested).
        uncertainties: Dictionary containing uncertainty values (same structure as analysis).
        top_n: Number of top fields to display.
        ascending: Whether to sort in ascending order.
    """
    # Flatten the dictionaries
    flattened_analysis = flatten_dict(analysis)
    if uncertainties:
        flattened_uncertainties = flatten_dict(uncertainties)
    else:
        uncertainties_list = None

    # Prepare data by matching fields
    fields = list(flattened_analysis.keys())
    similarities = [flattened_analysis[field] for field in fields]

    if uncertainties:
        uncertainties_list = [flattened_uncertainties.get(field, None) for field in fields]

    # Create a DataFrame
    df = pd.DataFrame(
        {
            "field": fields,
            "similarity": similarities,
        }
    )

    if uncertainties:
        df["uncertainty"] = uncertainties_list

    # Sort by similarity and select top N
    df = df.sort_values(by="similarity", ascending=ascending).head(top_n)

    # Calculate layout dimensions
    label_width = max(len(field) for field in df["field"]) + 2  # Padding for alignment
    terminal_width = shutil.get_terminal_size().columns
    bar_width = terminal_width - label_width - 3  # Space for '| ' and extra padding

    # Determine scaling factor based on maximum similarity
    max_similarity = df["similarity"].max()
    scale = bar_width / max_similarity if max_similarity > 0 else 1

    # Generate and print bars
    for index, row in df.iterrows():
        field = row["field"]
        similarity = row["similarity"]
        if uncertainties:
            uncertainty = row["uncertainty"]
        else:
            uncertainty = None

        if similarity is None:
            continue  # Skip fields with no similarity value

        # Calculate bar length and uncertainty range
        bar_len = round(similarity * scale)
        if uncertainty is not None and uncertainty > 0:
            uncertainty_start = max(0, round((similarity - uncertainty) * scale))
            uncertainty_end = min(bar_width, round((similarity + uncertainty) * scale))
        else:
            uncertainty_start = bar_len
            uncertainty_end = bar_len  # No uncertainty to display

        # Build the bar string
        bar_string = ''
        for i in range(bar_width):
            if i < bar_len:
                if i < uncertainty_start:
                    char = '█'  # Solid block for certain part
                else:
                    char = '█'  # Lighter block for uncertainty overlap
            else:
                if i < uncertainty_end:
                    char = '░'  # Dash for upper uncertainty range
                else:
                    char = ' '  # Space for empty area
            bar_string += char

        # Print the label and bar
        score_field = f'[{similarity:.4f}]'

        print(f"{field:<{label_width}} {score_field} | {bar_string}")



import datetime
from typing import Any, Literal, Optional

import nanoid  # type: ignore
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from pydantic import BaseModel, Field, HttpUrl, field_serializer

from ..modalities import Modality


def scrapping_action(link: HttpUrl) -> dict[str, Any]:
    raise NotImplementedError("Scrapping action not implemented")


class CronSchedule(BaseModel):
    second: Optional[int] = Field(0, ge=0, le=59, description="Second (0-59), defaults to 0")
    minute: int = Field(..., ge=0, le=59, description="Minute (0-59)")
    hour: int = Field(..., ge=0, le=23, description="Hour (0-23)")
    day_of_month: Optional[int] = Field(None, ge=1, le=31, description="Day of the month (1-31), None means any day")
    month: Optional[int] = Field(None, ge=1, le=12, description="Month (1-12), None means every month")
    day_of_week: Optional[int] = Field(None, ge=0, le=6, description="Day of the week (0-6, Sunday = 0), None means any day")

    def to_cron_string(self) -> str:
        return f"{self.second or '*'} {self.minute} {self.hour} {self.day_of_month or '*'} {self.month or '*'} {self.day_of_week or '*'}"


from ..logs import AutomationConfig


class ScrappingConfig(AutomationConfig):
    object: Literal['automation.scrapping_cron'] = "automation.scrapping_cron"
    id: str = Field(default_factory=lambda: "scrapping_" + nanoid.generate(), description="Unique identifier for the scrapping job")

    # Scrapping Specific Config
    link: HttpUrl = Field(..., description="Link to be scrapped")
    schedule: CronSchedule

    updated_at: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(datetime.timezone.utc))

    # HTTP Config
    webhook_url: HttpUrl = Field(..., description="Url of the webhook to send the data to")
    webhook_headers: dict[str, str] = Field(default_factory=dict, description="Headers to send with the request")

    modality: Modality
    image_resolution_dpi: int = Field(default=96, description="Resolution of the image sent to the LLM")
    browser_canvas: Literal['A3', 'A4', 'A5'] = Field(default='A4', description="Sets the size of the browser canvas for rendering documents in browser-based processing. Choose a size that matches the document type.")

    # New attributes
    model: str = Field(..., description="Model used for chat completion")
    json_schema: dict[str, Any] = Field(..., description="JSON schema format used to validate the output data.")
    temperature: float = Field(default=0.0, description="Temperature for sampling. If not provided, the default temperature for the model will be used.", examples=[0.0])
    reasoning_effort: ChatCompletionReasoningEffort = Field(
        default="medium", description="The effort level for the model to reason about the input data. If not provided, the default reasoning effort for the model will be used."
    )

    @field_serializer('webhook_url', 'link')
    def url2str(self, val: HttpUrl) -> str:
        return str(val)



import copy
import datetime
import json
from typing import Any, List, Literal, Optional, Union

import nanoid  # type: ignore
from pydantic import BaseModel, Field, computed_field


from .._utils.json_schema import clean_schema, compute_schema_data_id
from .._utils.mime import generate_blake2b_hash_from_string
from .ai_models import Amount, LLMModel
from .jobs.base import InferenceSettings
from .mime import MIMEData


# Define the type alias for MetricType
MetricType = Literal["levenshtein", "jaccard", "hamming"]


# Define the structure for an individual item metric
class ItemMetric(BaseModel):
    id: str = Field(description="The ID of the item being measured")
    name: str = Field(description="The name of the item being measured")
    similarity: float = Field(description="The similarity score between 0 and 1")
    similarities: dict[str, Any] = Field(description="The similarity scores for each item in the list")
    flat_similarities: dict[str, Optional[float]] = Field(description="The similarity scores for each item in the list in dot notation format")
    aligned_similarity: float = Field(description="The similarity score between 0 and 1, after alignment")
    aligned_similarities: dict[str, Any] = Field(description="The similarity scores for each item in the list, after alignment")
    aligned_flat_similarities: dict[str, Optional[float]] = Field(description="The similarity scores for each item in the list in dot notation format, after alignment")


# Define the main MetricResult model
class MetricResult(BaseModel):
    item_metrics: List[ItemMetric] = Field(description="List of similarity metrics for individual items")
    mean_similarity: float = Field(description="The average similarity score across all items")
    aligned_mean_similarity: float = Field(description="The average similarity score across all items, after alignment")
    metric_type: MetricType = Field(description="The type of similarity metric used for comparison")


class DistancesResult(BaseModel):
    distances: dict[str, Any] = Field(description="List of distances for individual items")
    mean_distance: float = Field(description="The average distance across all items")
    metric_type: MetricType = Field(description="The type of distance metric used for comparison")


class PredictionMetadata(BaseModel):
    extraction_id: Optional[str] = Field(default=None, description="The ID of the extraction")
    likelihoods: Optional[dict[str, Any]] = Field(default=None, description="The likelihoods of the extraction")
    field_locations: Optional[dict[str, Any]] = Field(default=None, description="The field locations of the extraction")
    agentic_field_locations: Optional[dict[str, Any]] = Field(default=None, description="The field locations of the extraction extracted by an llm")
    consensus_details: Optional[list[dict[str, Any]]] = Field(default=None, description="The consensus details of the extraction")
    api_cost: Optional[Amount] = Field(default=None, description="The cost of the API call for this document (if any -- ground truth for example)")


class PredictionData(BaseModel):
    prediction: dict[str, Any] = Field(default={}, description="The result of the extraction or manual annotation")
    metadata: Optional[PredictionMetadata] = Field(default=None, description="The metadata of the prediction")


class Iteration(BaseModel):
    id: str = Field(default_factory=lambda: "eval_iter_" + nanoid.generate())
    inference_settings: InferenceSettings
    json_schema: dict[str, Any]
    predictions: list[PredictionData] = Field(default_factory=list, description="The predictions of the iteration for all the documents")
    metric_results: Optional[MetricResult] = Field(default=None, description="The metric results of the iteration")

    @computed_field  # type: ignore
    @property
    def schema_data_id(self) -> str:
        """Returns the SHA1 hash of the schema data, ignoring all prompt/description/default fields.

        Returns:
            str: A SHA1 hash string representing the schema data version.
        """
        return "sch_data_id_" + generate_blake2b_hash_from_string(
            json.dumps(
                clean_schema(
                    copy.deepcopy(self.json_schema),
                    remove_custom_fields=True,
                    fields_to_remove=["description", "default", "title", "required", "examples", "deprecated", "readOnly", "writeOnly"],
                ),
                sort_keys=True,
            ).strip()
        )

    # This is a computed field, it is exposed when serializing the object
    @computed_field  # type: ignore
    @property
    def schema_id(self) -> str:
        """Returns the SHA1 hash of the complete schema.

        Returns:
            str: A SHA1 hash string representing the complete schema version.
        """
        return "sch_id_" + generate_blake2b_hash_from_string(json.dumps(self.json_schema, sort_keys=True).strip())


class AnnotatedDocument(BaseModel):
    mime_data: MIMEData = Field(
        description="The mime data of the document. Can also be a BaseMIMEData, which is why we have this id field (to be able to identify the file, but id is equal to mime_data.id)"
    )
    annotation: dict[str, Any] = Field(default={}, description="The ground truth of the document")


class DocumentItem(AnnotatedDocument):
    annotation_metadata: Optional[PredictionMetadata] = Field(default=None, description="The metadata of the annotation when the annotation is a prediction")


class EvaluationDocument(DocumentItem):
    id: str = Field(description="The ID of the document. Equal to mime_data.id but robust to the case where mime_data is a BaseMIMEData")


class CreateIterationRequest(BaseModel):
    """
    Request model for performing a new iteration with custom inference settings and optional JSON schema.
    """

    inference_settings: InferenceSettings
    json_schema: Optional[dict[str, Any]] = None


class UpdateEvaluationDocumentRequest(BaseModel):
    annotation: Optional[dict[str, Any]] = Field(default=None, description="The ground truth of the document")
    annotation_metadata: Optional[PredictionMetadata] = Field(default=None, description="The metadata of the annotation when the annotation is a prediction")


class UpdateEvaluationRequest(BaseModel):
    name: Optional[str] = Field(default=None, description="The name of the document")
    documents: Optional[list[EvaluationDocument]] = Field(default=None, description="The documents of the evaluation")
    iterations: Optional[list[Iteration]] = Field(default=None, description="The iterations of the evaluation")
    json_schema: Optional[dict[str, Any]] = Field(default=None, description="The json schema of the evaluation")

    project_id: Optional[str] = Field(default=None, description="The ID of the project")

    @computed_field  # type: ignore
    @property
    def schema_data_id(self) -> Optional[str]:
        """Returns the SHA1 hash of the schema data, ignoring all prompt/description/default fields.

        Returns:
            str: A SHA1 hash string representing the schema data version.
        """
        if self.json_schema is None:
            return None

        return compute_schema_data_id(self.json_schema)

    # This is a computed field, it is exposed when serializing the object
    @computed_field  # type: ignore
    @property
    def schema_id(self) -> Optional[str]:
        """Returns the SHA1 hash of the complete schema.

        Returns:
            str: A SHA1 hash string representing the complete schema version.
        """
        if self.json_schema is None:
            return None
        return "sch_id_" + generate_blake2b_hash_from_string(json.dumps(self.json_schema, sort_keys=True).strip())


class Evaluation(BaseModel):
    id: str = Field(default_factory=lambda: "eval_" + nanoid.generate())
    updated_at: datetime.datetime = Field(default_factory=lambda: datetime.datetime.now(tz=datetime.timezone.utc))

    name: str
    documents: list[EvaluationDocument]
    iterations: list[Iteration]
    json_schema: dict[str, Any]

    project_id: str = Field(description="The ID of the project", default="default_spreadsheets")
    default_inference_settings: Optional[InferenceSettings] = Field(default=None, description="The default inference properties for the evaluation (mostly used in the frontend)")

    # @field_validator('iterations')
    # def validate_iterations_content_length(cls: Any, v: list[Iteration], values: Any) -> list[Iteration]:
    #     if 'ground_truth' in values:
    #         ground_truth_length = len(values['ground_truth'])
    #         for iteration in v:
    #             if len(iteration.content) != ground_truth_length:
    #                 raise ValueError(f"Iteration content length must match ground_truth length ({ground_truth_length})")
    #     return v

    @computed_field  # type: ignore
    @property
    def schema_data_id(self) -> str:
        """Returns the SHA1 hash of the schema data, ignoring all prompt/description/default fields.

        Returns:
            str: A SHA1 hash string representing the schema data version.
        """
        return compute_schema_data_id(self.json_schema)

    # This is a computed field, it is exposed when serializing the object
    @computed_field  # type: ignore
    @property
    def schema_id(self) -> str:
        """Returns the SHA1 hash of the complete schema.

        Returns:
            str: A SHA1 hash string representing the complete schema version.
        """
        return "sch_id_" + generate_blake2b_hash_from_string(json.dumps(self.json_schema, sort_keys=True).strip())


class AddIterationFromJsonlRequest(BaseModel):
    jsonl_gcs_path: str
